;*=====================================================================*/
;*    .../prgm/project/bigloo/api/fthread/src/Llib/fbackend.scm        */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Sun Feb 24 06:42:48 2008                          */
;*    Last change :  Sun Apr 20 15:58:30 2008 (serrano)                */
;*    Copyright   :  2008 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    The Fair Thread backend                                          */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_backend

   (include "_thread.sch")
   
   (import  __ft_%thread
	    __ft_%types
	    __ft_%scheduler
	    __ft_types
	    __ft_thread
            __ft_signal
	    __ft_scheduler)
   
   (static (class fthread-backend::thread-backend))
   
   (export (get-fthread-backend)))

;*---------------------------------------------------------------------*/
;*    *fthread-backend* ...                                            */
;*---------------------------------------------------------------------*/
(define *fthread-backend* #unspecified)

;*---------------------------------------------------------------------*/
;*    fthread-setup-backend! ...                                       */
;*---------------------------------------------------------------------*/
(define (fthread-setup-backend!)
   (set! *fthread-backend* (instantiate::fthread-backend (name "fthread")))
   (default-thread-backend-set! *fthread-backend*))

;*---------------------------------------------------------------------*/
;*    get-fthread-backend ...                                          */
;*---------------------------------------------------------------------*/
(define (get-fthread-backend)
   *fthread-backend*)

;*---------------------------------------------------------------------*/
;*    tb-make-thread ::fthread-backend ...                             */
;*---------------------------------------------------------------------*/
(define-method (tb-make-thread tb::fthread-backend thunk name)
   
   (define (execute-thread t)
      (current-thread-backend-set! (get-fthread-backend))
      (with-access::fthread t (%state %result cleanup scheduler)
	 ;; terminate is used to abruptly terminate a thread
	 (bind-exit (terminate)
	    (fthread-%terminate-set! t terminate)
	    (with-exception-handler
	       (lambda (e)
		  (let ((u (instantiate::uncaught-exception
			      (reason e))))
		     (fthread-%exc-result-set! t u)
		     (exception-notify e)
		     (terminate #f)))
	       (lambda ()
		  ;; store the result of the thread
		  (set! %result (thunk)))))
	 ;; broadcast a signal for the thread termination
	 (broadcast! (instantiate::%sigjoin (thread t)) %result)
	 ;; invoke the thread cleanup
	 (if (procedure? cleanup)
	     (if (correct-arity? cleanup 1)
		 (cleanup t)
		 (error t "Illegal cleanup function" cleanup)))
	 ;; kill the thread is now dead and switch back to the scheduler
	 (%thread-kill! t)))
   
   (if (not (correct-arity? thunk 0))
       (error 'make-thread
	      "Illegal thread body (should be a procedure of 0 arguments)"
	      thunk)
       (let ((the-name (cond
			  ((symbol? name)
			   name)
			  ((string? name)
			   (string->symbol name))
			  (else
			   (gensym 'fairthread)))))
	  (letrec ((t (instantiate::fthread
			 (%builtin ($fthread-new/name
				    (lambda () (execute-thread t))
				    the-name))
			 (name name))))
	     (%thread-setup! t)
	     t))))

;*---------------------------------------------------------------------*/
;*    tb-thread-sleep! ::fthread-backend ...                           */
;*---------------------------------------------------------------------*/
(define-method (tb-thread-sleep! tb::fthread-backend timeout)
   (let ((t (tb-current-thread tb)))
      (if (thread? t)
	  (when (>fx timeout 0) (%thread-timeout! t timeout))
	  (error 'thread-sleep! "No current thread" t))))

;*---------------------------------------------------------------------*/
;*    tb-current-thread ::fthread-backend ...                          */
;*---------------------------------------------------------------------*/
(define-method (tb-current-thread tb::fthread-backend)
   (if (not (scheduler? (current-scheduler)))
       #f
       (%scheduler-current-thread (current-scheduler))))

;*---------------------------------------------------------------------*/
;*    tb-thread-yield! ::fthread-backend ...                           */
;*---------------------------------------------------------------------*/
(define-method (tb-thread-yield! tb::fthread-backend)
   (let ((t (tb-current-thread tb)))
      (if (thread? t)
	  (%thread-yield! t)
	  (error 'thread-yield! "No current thread" t))))

;*---------------------------------------------------------------------*/
;*    FairThread initialization                                        */
;*---------------------------------------------------------------------*/
(fthread-setup-backend!)
