;*=====================================================================*/
;*    .../prgm/project/bigloo/api/gstreamer/examples/setmixer.scm      */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Jan  9 16:39:30 2008                          */
;*    Last change :  Thu Jan 31 10:54:07 2008 (serrano)                */
;*    Copyright   :  2008 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    Access the mixer of the local sound card.                        */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module gstreamer_example-setmixer
   (library gstreamer multimedia pthread)
   (main main))

;*---------------------------------------------------------------------*/
;*    channel ...                                                      */
;*---------------------------------------------------------------------*/
(define channel #f)
(define volume #f)

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main argv)
   (let loop ((args (cdr argv)))
      (cond
	 ((null? args)
	  (print "usage: setmixer [-V] channel val [channel val ...]")
	  (print "where val is either an integer or integer,integer."))
	 ((string=? (car argv) "-V")
	  (mixer-show))
	 ((null? (cdr args))
	  (mixer-show))
	 (else
	  (let ((channel (car args))
		(val (cadr args)))
	     (let ((m (pregexp-match "([0-9]+),([0-9]+)" val)))
		(if (pair? m)
		    (set-volume! channel
				 (vector (string->integer (cadr m))
					 (string->integer (caddr m))))
		    (set-volume! channel (string->integer val))))
	     (when (pair? (cddr args))
		(loop (cddr args))))))))

;*---------------------------------------------------------------------*/
;*    find-mixer ...                                                   */
;*---------------------------------------------------------------------*/
(define (find-mixer)
   (let loop ((fs (reverse (gst-registry-mixer-list))))
      (if (null? fs)
	  (error 'setmixer "Cannot find mixer in available plugins" #f)
	  (let ((mix (gst-element-factory-create (car fs) "mixer")))
	     (if (member "GstMixer" (gst-element-interface-list mix))
		 mix
		 (loop (cdr fs)))))))

;*---------------------------------------------------------------------*/
;*    mixer-show ...                                                   */
;*---------------------------------------------------------------------*/
(define (mixer-show)
   (let ((mix (find-mixer)))
      (tprint "mix=" (gst-element-element-factory mix))
      (gst-element-state-set! mix 'ready)
      (unwind-protect
	 (let ((tracks (gst-mixer-track-list mix)))
	    (for-each (lambda (t)
			 (let* ((lbl (gst-mixer-track-label t))
				(len (string-length lbl)))
			    (when (<fx len 20)
			       (display (make-string (-fx 20 len))))
			    (display lbl)
			    (display " - ")
			    (display (gst-mixer-volume mix t))
			    (printf " [~a-~a]\n"
				    (gst-mixer-track-min-volume t)
				    (gst-mixer-track-max-volume t))))
		      tracks))
	 (gst-element-state-set! mix 'null))))

;*---------------------------------------------------------------------*/
;*    set-volume! ...                                                  */
;*---------------------------------------------------------------------*/
(define (set-volume! channel volume)
   (let ((mix (find-mixer)))
      (gst-element-state-set! mix 'ready)
      (unwind-protect
	 (let ((tracks (gst-mixer-track-list mix)))
	    (for-each (lambda (t)
			 (when (string-ci=? (gst-mixer-track-label t) channel)
			    (gst-mixer-volume-set! mix t volume)))
		      tracks))
	 (gst-element-state-set! mix 'null))))
