/*=====================================================================*/
/*    .../project/bigloo/api/gstreamer/src/Clib/bglgst_plugin.c        */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Jan 11 11:29:03 2008                          */
/*    Last change :  Tue May 27 13:55:45 2008 (serrano)                */
/*    Copyright   :  2008 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    GSTREAMER plugin implementation                                  */
/*=====================================================================*/
#include <gst/gst.h>
#include "bglgst_config.h"
#if( BGL_GSTREAMER_HAVE_AUDIO )
#  include <gst/interfaces/mixer.h>
#endif
#include <string.h>
#include "bglgst.h"
#include "bglgst_plugin.h"

/*---------------------------------------------------------------------*/
/*    imports                                                          */
/*---------------------------------------------------------------------*/
#if( !BGL_GSTREAMER_HAVE_AUDIO )
extern obj_t make_vector( long, obj_t );
#  define GST_MIXER_TRACK( val ) ((void *)val)
#endif

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_gst_mixer_get_track ...                                      */
/*---------------------------------------------------------------------*/
obj_t
bgl_gst_mixer_get_track( GstElement *mixer, char *name, obj_t obj ) {
#if( BGL_GSTREAMER_HAVE_AUDIO )
   if( !G_TYPE_CHECK_INSTANCE_TYPE( G_OBJECT( mixer ), GST_TYPE_MIXER ) ) {
      C_SYSTEM_FAILURE( BGL_TYPE_ERROR,
			"gst-mixer-track-list",
			"gst-mixer expected",
			obj );
   } else {
      if( GST_IS_MIXER( mixer ) ) {
	 const GList *tracks = gst_mixer_list_tracks( GST_MIXER( mixer ) );

	 while( tracks ) {
	    GstMixerTrack *track = (GstMixerTrack *)(tracks->data);
	    if( !strcmp( track->label, name ) ) {
	       gst_object_ref( track );

	       return bgl_gst_mixer_track_new( track, BTRUE );

	       tracks = g_list_next( tracks );
	    } else {
	       C_SYSTEM_FAILURE( BGL_TYPE_ERROR,
				 "gst-mixer-track-list",
				 "gst-mixer not in ready-state",
				 obj );
	    }
	 }
      }
   }
#else   
   return BFALSE;
#endif   
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    mktrack ...                                                      */
/*---------------------------------------------------------------------*/
static void
mktrack( gpointer val, gpointer data ) {
   obj_t res = (obj_t)data;
   GstMixerTrack *track = GST_MIXER_TRACK( val );
   
   if( GST_IS_MIXER_TRACK( track ) && !GST_IS_MIXER_OPTIONS( track ) ) {
      obj_t btrack = bgl_gst_mixer_track_new( track, BTRUE );
      gst_object_ref( track );
      SET_CDR( res, MAKE_PAIR( btrack, CDR( res ) ) );
   }
}


/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_gst_mixer_track_list ...                                     */
/*---------------------------------------------------------------------*/
obj_t
bgl_gst_mixer_track_list( GstElement *mixer, obj_t obj ) {
#if( BGL_GSTREAMER_HAVE_AUDIO )
   if( !G_TYPE_CHECK_INSTANCE_TYPE( G_OBJECT( mixer ), GST_TYPE_MIXER ) ) {
      C_SYSTEM_FAILURE( BGL_TYPE_ERROR,
			"gst-mixer-track-list",
			"gst-mixer expected",
			obj );
   } else {
      if( GST_IS_MIXER( mixer ) ) {
	 const GList *tracks = gst_mixer_list_tracks( GST_MIXER( mixer ) );
	 obj_t res = MAKE_PAIR( BNIL, BNIL );
	 
	 if( tracks ) g_list_foreach( (GList *)tracks, mktrack, (gpointer)res );

	 return CDR( res );
      } else {
	 C_SYSTEM_FAILURE( BGL_TYPE_ERROR,
			   "gst-mixer-track-list",
			   "gst-mixer not in ready-state",
			   obj );
      }
   }
#else   
   return BNIL;
#endif   
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_gst_mixer_get_volume ...                                     */
/*---------------------------------------------------------------------*/
obj_t
bgl_gst_mixer_get_volume( GstElement *mix, GstMixerTrack *track ) {
#if( BGL_GSTREAMER_HAVE_AUDIO )
   int num = track->num_channels;
   int *buf = alloca( sizeof( gint ) * num );
   obj_t res = create_vector( num );

   gst_mixer_get_volume( GST_MIXER( mix ), track, buf );

   while( num-- > 0 ) {
      VECTOR_SET( res, num, BINT( buf[ num ] ) );
   }

   return res;
#else   
   return make_vector( 0, BINT( 0 ) );
#endif   
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_gst_mixer_set_volume ...                                     */
/*---------------------------------------------------------------------*/
obj_t
bgl_gst_mixer_set_volume( GstElement *mix, GstMixerTrack *track, obj_t obj ) {
#if( BGL_GSTREAMER_HAVE_AUDIO )
   int num = track->num_channels;
   int *buf = alloca( sizeof( gint ) * num );

   if( INTEGERP( obj )  ) {
      while( num > 0 ) {
	 buf[ --num ] = CINT( obj );
      }
   } else if( VECTORP( obj ) ) {
      if( VECTOR_LENGTH( obj ) != num ) {
	 C_SYSTEM_FAILURE( BGL_ERROR,
			   "gst-mixer-volume-set!",
			   "Wrong number of volume values",
			   obj );
      }
      
      while( num > 0 ) {
	 num--;
	 buf[ num ] = CINT( VECTOR_REF( obj, num ) );
      }
   } else {
      C_SYSTEM_FAILURE( BGL_TYPE_ERROR,
			"gst-mixer-volume-set!",
			"vector or integer expected",
			obj );
   }

   gst_mixer_set_volume( GST_MIXER( mix ), track, buf );
#endif
   return obj;
}
