/*=====================================================================*/
/*    .../prgm/project/bigloo/api/gstreamer/src/Plugin/gstbigloo.c     */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Feb 11 17:03:37 2008                          */
/*    Last change :  Tue Feb 12 16:02:05 2008 (serrano)                */
/*    Copyright   :  2008 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    Bigloo proc plugin.                                              */
/*=====================================================================*/
#include <bigloo.h>
#include <gst/gst.h>
#include "bglgst_config.h"
#include "gstbigloo.h"

/*---------------------------------------------------------------------*/
/*    Plugin configuration                                             */
/*---------------------------------------------------------------------*/
#define PLUGIN_VERSION BGL_RELEASE_NUMBER
#define PLUGIN_PACKAGE "Bigloo"
#define PLUGIN_LICENSE "GPL"
#define PLUGIN_URL "http://www.inria.fr/mimosa/fp/Bigloo"

#define PACKAGE PLUGIN_PACKAGE

/*---------------------------------------------------------------------*/
/*    Plugin data structures                                           */
/*---------------------------------------------------------------------*/
GST_DEBUG_CATEGORY_STATIC ( gst_bigloo_debug );
#define GST_CAT_DEFAULT gst_bigloo_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT,
  ARG_PROCEDURE,
};

/*---------------------------------------------------------------------*/
/*    Plugin's pad                                                     */
/*---------------------------------------------------------------------*/
static GstStaticPadTemplate sink_factory =
   GST_STATIC_PAD_TEMPLATE( "sink",
			    GST_PAD_SINK,
			    GST_PAD_ALWAYS,
			    GST_STATIC_CAPS( "ANY" ) );

static GstStaticPadTemplate src_factory =
   GST_STATIC_PAD_TEMPLATE( "src",
			    GST_PAD_SRC,
			    GST_PAD_ALWAYS,
			    GST_STATIC_CAPS( "ANY" ) );

/*---------------------------------------------------------------------*/
/*    Plugin boilerplate                                               */
/*---------------------------------------------------------------------*/
GST_BOILERPLATE( Gstbigloo, gst_bigloo, GstElement, GST_TYPE_ELEMENT );

// should use GstBaseTransform

/*---------------------------------------------------------------------*/
/*    Static declarations                                              */
/*---------------------------------------------------------------------*/
static void gst_bigloo_set_property( GObject * object,
				     guint prop_id,
				     const GValue * value,
				     GParamSpec * pspec);
static void gst_bigloo_get_property( GObject * object,
				     guint prop_id,
				     GValue * value,
				     GParamSpec * pspec );

static gboolean gst_bigloo_set_caps( GstPad * pad, GstCaps * caps );
static GstFlowReturn gst_bigloo_chain( GstPad * pad, GstBuffer * buf );

static GstStateChangeReturn gst_bigloo_change_state( GstElement *element,
						     GstStateChange transition );

static void gst_bigloo_finalize( GObject *o );

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_base_init ...                                         */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_base_init( gpointer gclass ) {
   static GstElementDetails element_details = {
      "Bigloo plugin",
      "Generic/PluginTemplate",
      "Generic Template Element",
      "Manuel Serrano <Manuel.Serrano@inria.fr>"
   };
   GstElementClass *element_class = GST_ELEMENT_CLASS( gclass );

   gst_element_class_add_pad_template(
      element_class, gst_static_pad_template_get( &src_factory ) );
   gst_element_class_add_pad_template(
      element_class, gst_static_pad_template_get( &sink_factory ) );
   gst_element_class_set_details(
      element_class, &element_details );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_class_init ...                                        */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_class_init( GstbiglooClass *klass ) {
   GObjectClass *gobject_class;
   GstElementClass *gstelement_class;

   gobject_class = (GObjectClass *)klass;
   gstelement_class = (GstElementClass *)klass;

   gobject_class->set_property = gst_bigloo_set_property;
   gobject_class->get_property = gst_bigloo_get_property;
   gobject_class->finalize = gst_bigloo_finalize;

   gstelement_class->change_state = gst_bigloo_change_state;
  
   g_object_class_install_property(
      gobject_class, ARG_SILENT,
      g_param_spec_boolean( "silent", "Silent",
			    "Produce verbose output ?",
			    FALSE, G_PARAM_READWRITE ) );
   g_object_class_install_property(
      gobject_class, ARG_PROCEDURE,
      g_param_spec_boolean( "procedure", "procedure",
			    "The Bigloo procedure associated with the filter",
			    FALSE, G_PARAM_READWRITE ) );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_init ...                                              */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_init( Gstbigloo *filter, GstbiglooClass *gclass ) {
   GstElementClass *klass = GST_ELEMENT_GET_CLASS( filter );

   fprintf( stderr, "gst_bigloo_init: %p\n", filter );
   
   filter->sinkpad =
      gst_pad_new_from_template( gst_element_class_get_pad_template( klass, "sink"),
				 "sink" );
   gst_pad_set_setcaps_function( filter->sinkpad,
				 GST_DEBUG_FUNCPTR( gst_bigloo_set_caps ) );
   gst_pad_set_getcaps_function( filter->sinkpad,
				 GST_DEBUG_FUNCPTR( gst_pad_proxy_getcaps ) );

   filter->srcpad =
      gst_pad_new_from_template( gst_element_class_get_pad_template( klass, "src"),
				 "src" );
   gst_pad_set_getcaps_function( filter->srcpad,
				 GST_DEBUG_FUNCPTR( gst_pad_proxy_getcaps ) );

   gst_element_add_pad( GST_ELEMENT( filter), filter->sinkpad );
   gst_element_add_pad( GST_ELEMENT( filter), filter->srcpad );
   gst_pad_set_chain_function( filter->sinkpad,
			       GST_DEBUG_FUNCPTR( gst_bigloo_chain ) );
   filter->silent = FALSE;
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    gst_bigloo_id ...                                                */
/*---------------------------------------------------------------------*/
static obj_t
gst_bigloo_id( void *obj, obj_t flag ) {
   return obj;
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_finalize ...                                          */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_finalize( GObject *o ) {
   fprintf( stderr, "finalize: %p\n", o );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_set_property ...                                      */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_set_property( GObject *object,
			 guint prop_id,
			 const GValue *value,
			 GParamSpec *pspec ) {
  Gstbigloo *filter = GST_BIGLOO( object );

  switch( prop_id ) {
     case ARG_SILENT:
	filter->silent = g_value_get_boolean( value );
	break;
     case ARG_PROCEDURE:
	filter->procedure = (obj_t)g_value_get_pointer( value );
	break;
     default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID( object, prop_id, pspec );
	break;
  }
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gst_bigloo_get_property ...                                      */
/*---------------------------------------------------------------------*/
static void
gst_bigloo_get_property( GObject *object,
			 guint prop_id,
			 GValue *value,
			 GParamSpec *pspec ) {
  Gstbigloo *filter = GST_BIGLOO( object );

  switch( prop_id ) {
     case ARG_SILENT:
	g_value_set_boolean( value, filter->silent );
	break;
     case ARG_PROCEDURE:
	g_value_set_pointer( value, filter->procedure );
	break;
     default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID( object, prop_id, pspec );
	break;
  }
}

/*---------------------------------------------------------------------*/
/*    static gboolean                                                  */
/*    gst_bigloo_set_caps ...                                          */
/*---------------------------------------------------------------------*/
static gboolean
gst_bigloo_set_caps( GstPad * pad, GstCaps * caps ) {
   Gstbigloo *filter;
   GstPad *otherpad;

   filter = GST_BIGLOO( gst_pad_get_parent( pad ) );
   otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

   return gst_pad_set_caps( pad, caps );
}

/*---------------------------------------------------------------------*/
/*    static GstFlowReturn                                             */
/*    gst_bigloo_chain ...                                             */
/*---------------------------------------------------------------------*/
static GstFlowReturn
gst_bigloo_chain( GstPad *pad, GstBuffer *buf ) {
   Gstbigloo *filter;

   filter = GST_BIGLOO( GST_OBJECT_PARENT( pad ) );

   if( filter->silent == FALSE )
      g_print( "I'm plugged, therefore I'm in.\n" );

   return gst_pad_push( filter->srcpad, buf );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    gst_bigloo_allocate_memory ...                                   */
/*---------------------------------------------------------------------*/
static int
gst_bigloo_allocate_memory( Gstbigloo *filter ) {
   return 1;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_bigloo_free_memory ...                                       */
/*---------------------------------------------------------------------*/
static int
gst_bigloo_free_memory( Gstbigloo *filter ) {
   return 1;
}

/*---------------------------------------------------------------------*/
/*    static GstStateChangeReturn                                      */
/*    gst_bigloo_change_state ...                                      */
/*---------------------------------------------------------------------*/
static GstStateChangeReturn
gst_bigloo_change_state( GstElement *element, GstStateChange transition ) {
   GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
   Gstbigloo *filter = GST_BIGLOO( element );

   switch( transition ) {
      case GST_STATE_CHANGE_NULL_TO_READY:
	 if( !gst_bigloo_allocate_memory( filter ) )
	    return GST_STATE_CHANGE_FAILURE;
	 break;
      default:
	 break;
   }
  
   ret = GST_ELEMENT_CLASS( parent_class )->change_state( element, transition );
   if( ret == GST_STATE_CHANGE_FAILURE )
      return ret;

   switch( transition ) {
      case GST_STATE_CHANGE_READY_TO_NULL:
	 gst_bigloo_free_memory( filter );
	 break;
      default:
	 break;
   }

   return ret;
}

/*---------------------------------------------------------------------*/
/*    static gboolean                                                  */
/*    plugin_init ...                                                  */
/*---------------------------------------------------------------------*/
static gboolean
plugin_init( GstPlugin * plugin ) {
   GST_DEBUG_CATEGORY_INIT( gst_bigloo_debug, "bigloo-proc",
			    0, "Bigloo procedure plugin" );

   return gst_element_register( plugin,
				"bigloo",
				GST_RANK_NONE,
				GST_TYPE_BIGLOO );
}

/*---------------------------------------------------------------------*/
/*    GST_PLUGIN_DEFINE ...                                            */
/*---------------------------------------------------------------------*/
GST_PLUGIN_DEFINE( GST_VERSION_MAJOR,
		   GST_VERSION_MINOR,
		   "bigloo",
		   "Bigloo plugin",
		   plugin_init,
		   PLUGIN_VERSION,
		   PLUGIN_LICENSE,
		   PLUGIN_PACKAGE,
		   PLUGIN_URL )

/*---------------------------------------------------------------------*/
/*    gboolean                                                         */
/*    bgl_plugin_proc_init ...                                         */
/*---------------------------------------------------------------------*/
gboolean
bgl_plugin_proc_init() {
#if( defined( GST_PLUGIN_DEFINE_STATIC ) )
   _gst_plugin_register_static( &gst_plugin_desc );
#else      
   gst_plugin_register_static( GST_VERSION_MAJOR,
			       GST_VERSION_MINOR,
			       "bigloo",
			       "Bigloo plugin",
			       plugin_init, PLUGIN_VERSION,
			       PLUGIN_LICENSE,
			       PLUGIN_PACKAGE,
			       PLUGIN_URL );
#endif
}

