;*=====================================================================*/
;*    .../project/bigloo/api/multimedia/examples/musicplay.scm         */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Jan 31 11:44:28 2008                          */
;*    Last change :  Wed Feb  6 15:03:02 2008 (serrano)                */
;*    Copyright   :  2008 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    A simple music player                                            */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module musicplay
   (library pthread multimedia)
   (main main)
   (eval (export m)))

;*---------------------------------------------------------------------*/
;*    m ...                                                            */
;*---------------------------------------------------------------------*/
(define m #f)

;*---------------------------------------------------------------------*/
;*    onchange ...                                                     */
;*---------------------------------------------------------------------*/
(define (onchange key val player)
   (case key
      ((volume)
       (print ";; volume set to " val))
      ((state)
       (print ";; " val "...")
       (when (eq? val 'play)
	  (print ";; meta " (music-meta player))
	  (print ";; properties " (music-properties player))))))

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main args)
   (eval '(library-load 'multimedia))
   (let ((files '())
	 (backend 'mpg123)
	 (command #f))
      (args-parse (cdr args)
	 (("--mpg123" (help "Select the mpg123 back-end"))
	  (set! backend 'mpg123))
	 (("--mpc" (help "Select the mpc back-end"))
	  (set! backend 'mpc))
	 (("--mplayer" (help "Select the mplayer back-end"))
	  (set! backend 'mplayer))
	 (("--command" ?cmd (help "Set the command path"))
	  (set! command cmd))
	 (("--help" (help "This help"))
	  (print "usage: music [options] file ...")
	  (args-parse-usage #f)
	  (exit 0))
	 (else
	  (set! files (cons else files))))
      (print "Welcome to musicplay (backend: " backend ")...")
      (print "The music object is bound to the variable `m'.")
      (set! m
	    (case backend
	       ((mpg123)
		(if command
		    (instantiate::mpg123
		       (path command)
		       (onchange onchange)
		       (make-thread make-thread)
		       (thread-start thread-start-joinable!))
		    (instantiate::mpg123
		       (make-thread make-thread)
		       (onchange onchange)
		       (thread-start thread-start-joinable!))))
	       ((mplayer)
		(if command
		    (instantiate::mplayer
		       (path command)
		       (make-thread make-thread)
		       (onchange onchange)
		       (thread-start thread-start-joinable!))
		    (instantiate::mplayer
		       (make-thread make-thread)
		       (onchange onchange)
		       (thread-start thread-start-joinable!))))
	       ((mpc)
		(instantiate::mpc
		   (onchange onchange)))))
      (for-each (lambda (p) (music-playlist-add! m p)) (reverse files))
      (music-play m)
      (repl)))
