;*=====================================================================*/
;*    serrano/prgm/project/bigloo/api/multimedia/src/Llib/mpc.scm      */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Sat Jul 30 16:23:00 2005                          */
;*    Last change :  Sat May 24 10:29:07 2008 (serrano)                */
;*    Copyright   :  2005-08 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    MPC implementation                                               */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __multimedia-mpc

   (import __multimedia-music
	   __multimedia-music-event-loop)

   (export (class mpc::music
	      
	      (hello read-only (default #f))
	      (host read-only (default "localhost"))
	      (port read-only (default 6600))
	      (timeout read-only (default 10008993))
	      (prefix (default #f))

	      (%close::bool (default #f))
	      (%mpd (default #f))
	      (%socket (default #f)))))

;*---------------------------------------------------------------------*/
;*    mutex-locked? ...                                                */
;*---------------------------------------------------------------------*/
(define (mutex-locked? m)
   (let ((th (current-thread))
	 (st (mutex-state m)))
      (or (and (not th) (eq? st #unspecified))
	  (and (thread? th) (eq? (mutex-state m) th)))))

;*---------------------------------------------------------------------*/
;*    mpc-skip-line ...                                                */
;*---------------------------------------------------------------------*/
(define (mpc-skip-line ip)
   (let ((grammar (regular-grammar ((xall (or (out #\Newline #\Return) #a000)))
		     ((: (+ xall) (or #\Newline #\Return))
		      #t)
		     ((: (+ xall) #\Return #\Newline)
		      #t)
		     ((+ xall)
		      #t)
		     ((or #\Newline #\Return (: #\Return #\Newline))
		      #t)
		     (else
		      (the-failure)))))
      (read/rp grammar ip)))

;*---------------------------------------------------------------------*/
;*    mpc-connect! ...                                                 */
;*---------------------------------------------------------------------*/
(define (mpc-connect! mpc::mpc cmd)
   (with-access::mpc mpc (hello host port timeout %socket)
      [assert (mpc) (mutex-locked? (mpc-%mutex mpc))]
      
      (define (ack)
	 (when (string? hello)
	    (let ((po (socket-output %socket)))
	       (display hello po)
	       (flush-output-port po)))
	 (let ((pi (socket-input %socket)))
	    (let ((l (read-line pi)))
	       (unless (and (string? l) (substring-at? l "OK MPD" 0))
		  (raise
		   (instantiate::&io-port-error
		      (proc cmd)
		      (msg "Illegal MPC acknowledge")
		      (obj l))))
	       (substring l 6 (string-length l)))))
      
      (if (and (socket? %socket) (not (socket-down? %socket)))
	  (begin
	     ;; check if the connection is still up
	     (mpc-cmd mpc "ping")
	     (let ((l (mpc-skip-line (socket-input (mpc-%socket mpc)))))
		(when (eof-object? l)
		   (socket-close %socket)
		   (set! %socket (make-client-socket host port :timeout timeout))
		   (input-port-timeout-set! (socket-input %socket) timeout)
		   (mpc-%mpd-set! mpc (ack)))))
	  (begin
	     (set! %socket (make-client-socket host port :timeout timeout))
	     (input-port-timeout-set! (socket-input %socket) timeout)
	     (mpc-%mpd-set! mpc (ack))))))

;*---------------------------------------------------------------------*/
;*    mpc-connect/command! ...                                         */
;*---------------------------------------------------------------------*/
(define (mpc-connect/command! mpc::mpc status procname command parser)
   (with-access::mpc mpc (hello host port timeout %socket)
      [assert (mpc) (mutex-locked? (mpc-%mutex mpc))]
      
      (define (ack)
	 (when (string? hello)
	    (let ((po (socket-output %socket)))
	       (display hello po)
	       (flush-output-port po)))
	 (let ((pi (socket-input %socket)))
	    (let ((l (read-line pi)))
	       (unless (and (string? l) (substring-at? l "OK MPD" 0))
		  (raise
		   (instantiate::&io-port-error
		      (proc procname)
		      (msg "Illegal MPC acknowledge")
		      (obj l))))
	       (substring l 6 (string-length l)))))
      
      (define (execute armed)
	 ;; check if the connection is still up
	 (mpc-cmd mpc command)
	 (with-access::musicstatus (parser (socket-input %socket) status)
	       (state err)
	    (when (eq? state 'eof)
	       (socket-close %socket)
	       (set! %socket (make-client-socket host port :timeout timeout))
	       (input-port-timeout-set! (socket-input %socket) timeout)
	       (let ((a (ack)))
		  (if a
		      (begin
			 (mpc-%mpd-set! mpc a)
			 (if armed
			     (begin
				(set! state 'error)
				(set! err "Cannot establish connection"))
			     (execute #t)))
		      (begin
			 (set! state 'error)
			 (set! err "Illegal MPC acknowledge")))))))
      
      (if (and (socket? %socket) (not (socket-down? %socket)))
	  (execute #f)
	  (begin
	     (set! %socket (make-client-socket host port :timeout timeout))
	     (input-port-timeout-set! (socket-input %socket) timeout)
	     (let ((a (ack)))
		(if a
		    (begin
		       (mpc-%mpd-set! mpc a)
		       (execute #f))
		    (begin
		       (musicstatus-err-set! status "Illegal MPC acknowledge")
		       (musicstatus-state-set! status 'error))))))))

;*---------------------------------------------------------------------*/
;*    parse-error-msg ...                                              */
;*---------------------------------------------------------------------*/
(define (parse-error-msg fail port)
   (let ((s (string fail))
	 (l (read-line port)))
      (if (string? l)
	  (string-append "{" (string fail) "}" l)
	  fail)))

;*---------------------------------------------------------------------*/
;*    *mpc-string-grammar* ...                                         */
;*---------------------------------------------------------------------*/
(define *mpc-string-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (out " \r\n") (+ (out "\n")) #\Newline)
       (the-substring 0 -1))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal string value")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-symbol-grammar* ...                                         */
;*---------------------------------------------------------------------*/
(define *mpc-symbol-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (out " \r\n") (+ (out "\n")) #\Newline)
       (the-subsymbol 0 -1))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal string value")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-ignore-grammar* ...                                         */
;*---------------------------------------------------------------------*/
(define *mpc-ignore-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (* all) #\Newline)
       #unspecified)
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal ignore value")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-value-grammar* ...                                          */
;*---------------------------------------------------------------------*/
(define *mpc-value-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (* all) #\Newline)
       (the-substring 0 -1))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal ignore value")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-integer-grammar* ...                                        */
;*---------------------------------------------------------------------*/
(define *mpc-integer-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (+ digit) #\Newline)
       (the-fixnum))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal integer")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-time-grammar* ...                                           */
;*---------------------------------------------------------------------*/
(define *mpc-time-grammar*
   (regular-grammar (status)
      ((* blank)
       (ignore))
      ((: (+ digit) ":")
       (let* ((tm (the-fixnum))
	      (du (read/rp *mpc-integer-grammar* (the-port))))
	  (musicstatus-songpos-set! status tm)
	  (musicstatus-songlength-set! status du)))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal integer")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-ellapsed-grammar* ...                                       */
;*---------------------------------------------------------------------*/
(define *mpc-ellapsed-grammar*
   (regular-grammar ()
      ((* blank)
       (ignore))
      ((: (+ digit) ":")
       (let ((v (the-fixnum)))
	  (read/rp *mpc-integer-grammar* (the-port))
	  v))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc)
	   (msg "Illegal integer")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-playlist-grammar* ...                                       */
;*---------------------------------------------------------------------*/
(define *mpc-playlist-grammar*
   (regular-grammar ()
      ((: (+ digit) #\:)
       (read/rp *mpc-string-grammar* (the-port)))
      ("OK\n"
       'ok)
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc-playlist-get)
	   (msg "Illegal playlist value")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-song-grammar* ...                                           */
;*---------------------------------------------------------------------*/
(define *mpc-song-grammar*
   (regular-grammar ()
      ((: (+ alpha) #\:)
       (let ((k (the-keyword)))
	  (case k
	     ((file:)
	      (cons k (read/rp *mpc-string-grammar* (the-port))))
	     ((Title:)
	      (cons title: (read/rp *mpc-string-grammar* (the-port))))
	     ((Name:)
	      (cons name: (read/rp *mpc-string-grammar* (the-port))))
	     ((Id:)
	      (cons id: (read/rp *mpc-integer-grammar* (the-port))))
	     ((Pos:)
	      (cons pos: (read/rp *mpc-integer-grammar* (the-port))))
	     ((Track:)
	      (cons track: (read/rp *mpc-integer-grammar* (the-port))))
	     ((Time:)
	      (cons time: (read/rp *mpc-integer-grammar* (the-port))))
	     (else
	      (read/rp *mpc-ignore-grammar* (the-port))
	      (ignore)))))
      ("OK\n"
       'ok)
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc-song)
	   (msg "Illegal song status")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-khz-grammar* ...                                            */
;*---------------------------------------------------------------------*/
(define *mpc-khz-grammar*
   (regular-grammar ()
      ((+ (in " \t"))
       (ignore))
      ((+ digit)
       (let ((n (the-fixnum)))
	  (read/rp *mpc-ignore-grammar* (the-port))
	  n))
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc-song)
	   (msg "Illegal khz")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-status-grammar* ...                                         */
;*---------------------------------------------------------------------*/
(define *mpc-status-grammar*
   (regular-grammar (status)
      ("OK\n"
       status)
      ((: "s" (+ alpha) "g" #\:)
       ;; song
       (let ((k (the-keyword)))
	  (if (eq? k song:)
	      (let ((song (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-song-set! status song))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "s" (+ alpha) "d" #\:)
       ;; songid
       (let ((k (the-keyword)))
	  (if (eq? k songid:)
	      (let ((songid (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-songid-set! status songid))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "st" (+ alpha) #\:)
       ;; state
       (let ((k (the-keyword)))
	  (if (eq? k state:)
	      (let ((state (read/rp *mpc-symbol-grammar* (the-port))))
		 (musicstatus-state-set! status state))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "t" (+ alpha) #\:)
       ;; time
       (let ((k (the-keyword)))
	  (if (eq? k time:)
	      (read/rp *mpc-time-grammar* (the-port) status)
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "v" (+ alpha) #\:)
       ;; volume
       (let ((k (the-keyword)))
	  (if (eq? k volume:)
	      (let ((volume (read/rp *mpc-integer-grammar* (the-port)))) 
		 (musicstatus-volume-set! status volume))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "p" (+ alpha) "h" #\:)
       ;; playlist and playlistlength
       (let ((k (the-keyword)))
	  (if (eq? k playlistlength:)
	      (let ((plen (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-playlistlength-set! status plen))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "p" (+ alpha) "t" #\:)
       ;; playlist and playlistlength
       (let ((k (the-keyword)))
	  (if (eq? k playlist:)
	      (let ((pid (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-playlistid-set! status pid))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "e" (+ alpha) #\:)
       ;; error
       (let ((k (the-keyword)))
	  (if (eq? k error:)
	      (let ((err (read/rp *mpc-value-grammar* (the-port))))
		 (musicstatus-err-set! status err))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "b" (+ alpha) #\:)
       ;; bitrate
       (let ((k (the-keyword)))
	  (if (eq? k bitrate:)
	      (let ((bitrate (read (the-port))))
		 (musicstatus-bitrate-set! status bitrate))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "a" (+ alpha) #\:)
       ;; audio
       (let ((k (the-keyword)))
	  (if (eq? k audio:)
	      (let ((khz (read/rp *mpc-khz-grammar* (the-port))))
		 (musicstatus-khz-set! status khz))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "x" (+ alpha) #\:)
       ;; xfade
       (let ((k (the-keyword)))
	  (if (eq? k xfade:)
	      (let ((xfade (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-xfade-set! status xfade))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "ra" (+ alpha) #\:)
       ;; randgom
       (let ((k (the-keyword)))
	  (if (eq? k random:)
	      (let ((random (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-random-set! status (=fx random 1)))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: "re" (+ alpha) #\:)
       ;; repeat
       (let ((k (the-keyword)))
	  (if (eq? k repeat:)
	      (let ((repeat (read/rp *mpc-integer-grammar* (the-port))))
		 (musicstatus-repeat-set! status (=fx repeat 1)))
	      (read/rp *mpc-ignore-grammar* (the-port)))
	  (ignore)))
      ((: (+ alpha) #\:)
       (read/rp *mpc-ignore-grammar* (the-port))
       (ignore))
      ((: "ACK " (+ all))
       (musicstatus-err-set! status (the-string))
       (ignore))
      ((+ #\Newline)
       (ignore))
      (else
       ;; a parsing error
       (musicstatus-song-set! status 0)
       (musicstatus-songid-set! status 0)
       (musicstatus-playlistlength-set! status 0)
       (musicstatus-songpos-set! status 0)
       (musicstatus-songlength-set! status 0)
       (musicstatus-khz-set! status 0)
       (musicstatus-bitrate-set! status 0)
       (if (eof-object? (the-failure))
	   (begin
	      (musicstatus-state-set! status 'eof)
	      (unless (string? (musicstatus-err status))
		 (musicstatus-err-set! status "Illegal end-of-file")))
	   (begin
	      (musicstatus-state-set! status 'error)
	      (unless (string? (musicstatus-err status))
		 (musicstatus-err-set! status "Parsing error")))))))

;*---------------------------------------------------------------------*/
;*    *mpc-status-time-grammar* ...                                    */
;*---------------------------------------------------------------------*/
(define *mpc-status-time-grammar*
   (regular-grammar ()
      ((: (+ alpha) #\:)
       (let ((k (the-keyword)))
	  (case k
	     ((time:)
	      (read/rp *mpc-ellapsed-grammar* (the-port)))
	     (else
	      (read/rp *mpc-ignore-grammar* (the-port))
	      (ignore)))))
      ("OK\n"
       'ok)
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc-time)
	   (msg "Illegal status")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    *mpc-status-volume-grammar* ...                                  */
;*---------------------------------------------------------------------*/
(define *mpc-status-volume-grammar*
   (regular-grammar ()
      ((: (+ alpha) #\:)
       (let ((k (the-keyword)))
	  (case k
	     ((volume:)
	      (read/rp *mpc-integer-grammar* (the-port)))
	     (else
	      (read/rp *mpc-ignore-grammar* (the-port))
	      (ignore)))))
      ("OK\n"
       'ok)
      (else
       (raise
	(instantiate::&io-parse-error
	   (proc 'mpc-volume)
	   (msg "Illegal status")
	   (obj (parse-error-msg (the-failure) (the-port))))))))

;*---------------------------------------------------------------------*/
;*    mpc-ok ...                                                       */
;*---------------------------------------------------------------------*/
(define (mpc-ok mpc::mpc)
   [assert (mpc) (mutex-locked? (mpc-%mutex mpc))]
   (let ((l (read-line (socket-input (mpc-%socket mpc)))))
      (and (string? l) (substring-at? l "OK" 0))))
   
;*---------------------------------------------------------------------*/
;*    mpc-cmd ...                                                      */
;*---------------------------------------------------------------------*/
(define (mpc-cmd mpc::mpc cmd::bstring)
   [assert (mpc cmd) (mutex-locked? (mpc-%mutex mpc))]
   (let ((po (socket-output (mpc-%socket mpc))))
      (with-handler
	 (lambda (e)
	    (error-notify e)
	    (display "command ignored...\n" (current-error-port)))
	 (begin
	    (display-string cmd po)
	    (newline po)
	    (flush-output-port po)))))
    
;*---------------------------------------------------------------------*/
;*    music-close ::mpc ...                                            */
;*---------------------------------------------------------------------*/
(define-method (music-close mpc::mpc)
   ;; abort the event loop
   (music-event-loop-abort! mpc)
   (with-access::mpc mpc (%socket %close %mutex)
      (with-lock %mutex
	 (lambda ()
	    (unless %close
	       (set! %close #t)
	       (when (and (socket? %socket) (not (socket-down? %socket)))
		  (mpc-cmd mpc "ping")
		  (let ((l (read-line (socket-input (mpc-%socket mpc)))))
		     ;; issue a close command
		     (unless (eof-object? l) (mpc-cmd mpc "close"))
		     ;; close the socket
		     (socket-close %socket)))
	       (set! %socket #f))))))

;*---------------------------------------------------------------------*/
;*    music-closed? ::mpc ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-closed? mpc::mpc)
   (with-access::mpc mpc (%close)
      %close))
   
;*---------------------------------------------------------------------*/
;*    music-reset! ::mpc ...                                           */
;*---------------------------------------------------------------------*/
(define-method (music-reset! mpc::mpc)
   (with-access::mpc mpc (%socket %mutex)
      (with-lock %mutex
	 (lambda ()
	    (when (and (socket? %socket) (not (socket-down? %socket)))
	       (socket-close %socket)
	       (set! %socket #f))))))
   
;*---------------------------------------------------------------------*/
;*    music-playlist-get ::mpc ...                                     */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-get mpc::mpc)
   
   (define (qualify-path s)
      (if (or (substring-at? s "http:" 0) (substring-at? s "https:" 0))
	  s
	  (string-append (mpc-prefix mpc) s)))
   
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-playlist-get)
	 (mpc-cmd mpc "playlist")
	 (let ((ip (socket-input (mpc-%socket mpc))))
	    (let loop ((ser '()))
	       (let ((l (read/rp *mpc-playlist-grammar* ip)))
		  (if (eq? l 'ok)
		      (map! qualify-path (reverse! ser))
		      (loop (cons l ser)))))))))

;*---------------------------------------------------------------------*/
;*    music-playlist-add! ::mpc ...                                    */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-add! mpc::mpc s)
   (with-access::mpc mpc (%mutex prefix)
      (with-lock (mpc-%mutex mpc)
	 (lambda ()
	    (mpc-connect! mpc 'music-playlist-add!)
	    (let ((s (if (and (string? prefix) (substring-at? s prefix 0))
			 (substring s (string-length prefix) (string-length s))
			 s)))
	       (mpc-cmd mpc (string-append "add \"" s "\"")))
	    (mpc-ok mpc)))))

;*---------------------------------------------------------------------*/
;*    music-playlist-delete! ::mpc ...                                 */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-delete! mpc::mpc num)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-playlist-delete!)
	 (mpc-cmd mpc (string-append "delete " (integer->string num)))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-playlist-clear! ::mpc ...                                  */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-clear! mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-playlist-clear!)
	 (mpc-cmd mpc "clear")
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-update-status! ...                                         */
;*---------------------------------------------------------------------*/
(define-method (music-update-status! mpc::mpc status::musicstatus)
   
   (define (status-parser ip status)
      (with-access::musicstatus status (songpos songlength bitrate khz err)
	 ;; reset the properties that are optional in mpd status report
	 (set! songpos 0)
	 (set! songlength 0)
	 (set! bitrate 0)
	 (set! khz 0)
	 (set! err #f))
      (read/rp *mpc-status-grammar* ip status)
      status)
   
   (mutex-lock! (mpc-%mutex mpc))
   (unwind-protect
      (mpc-connect/command! mpc status 'music-status "status" status-parser)
      (mutex-unlock! (mpc-%mutex mpc))))

;*---------------------------------------------------------------------*/
;*    music-status ...                                                 */
;*---------------------------------------------------------------------*/
(define-method (music-status mpc::mpc)
   (let ((status (music-%status mpc)))
      (music-update-status! mpc status)
      status))

;*---------------------------------------------------------------------*/
;*    music-song ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-song mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-song)
	 (mpc-cmd mpc "currentsong")
	 (let ((ip (socket-input (mpc-%socket mpc))))
	    (let loop ((ser '()))
	       (let ((l (read/rp *mpc-song-grammar* ip)))
		  (if (eq? l 'ok)
		      (reverse! ser)
		      (loop (cons l ser)))))))))

;*---------------------------------------------------------------------*/
;*    music-time ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-songpos::int mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-time)
	 (mpc-cmd mpc "status")
	 (let ((ip (socket-input (mpc-%socket mpc))))
	    (let loop ((tm -1))
	       (let ((l (read/rp *mpc-status-time-grammar* ip)))
		  (cond
		     ((integer? l)
		      (loop l))
		     ((eq? l 'ok)
		      tm)
		     (else
		      (loop tm)))))))))

;*---------------------------------------------------------------------*/
;*    music-play ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-play mpc::mpc . song)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-play)
	 (mpc-cmd mpc (if (null? song) "play" (format "play ~a" (car song))))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-seek ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-seek mpc::mpc ntime::obj . nsong)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-seek)
	 (music-update-status! mpc (music-%status mpc))
	 (with-access::musicstatus (music-%status mpc) (song songpos)
	    (let ((song (if (null? nsong) song (car nsong)))
		  (time (if (fixnum? ntime)
			    ntime
			    (+ (if (pair? songpos) (car songpos) songpos)
			       (flonum->fixnum ntime)))))
	       (mpc-cmd mpc (format "seek ~a ~a" song time))))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-stop ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-stop mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-stop)
	 (mpc-cmd mpc "stop")
	 (mpc-ok mpc)
	 (mpc-cmd mpc "clearerror")
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-pause ::mpc ...                                            */
;*---------------------------------------------------------------------*/
(define-method (music-pause mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-pause)
	 (mpc-cmd mpc "pause")
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-next ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-next mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-next)
	 (mpc-cmd mpc "next")
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-prev ::mpc ...                                             */
;*---------------------------------------------------------------------*/
(define-method (music-prev mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-prev)
	 (mpc-cmd mpc "previous")
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-crossfade ::mpc ...                                        */
;*---------------------------------------------------------------------*/
(define-method (music-crossfade mpc::mpc sec::int)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-prev)
	 (mpc-cmd mpc (string-append "crossfade " (integer->string sec)))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-random-set! ::mpc ...                                      */
;*---------------------------------------------------------------------*/
(define-method (music-random-set! mpc::mpc flag::bool)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-prev)
	 (mpc-cmd mpc (if "random 1" "random 0" ))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-repeat-set! ::mpc ...                                      */
;*---------------------------------------------------------------------*/
(define-method (music-repeat-set! mpc::mpc flag::bool)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-prev)
	 (mpc-cmd mpc (if "repeat 1" "repeat 0" ))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-volume-get ::mpc ...                                       */
;*---------------------------------------------------------------------*/
(define-method (music-volume-get mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-volume-get)
	 (mpc-cmd mpc "status")
	 (let ((ip (socket-input (mpc-%socket mpc))))
	    (let loop ((tm -1))
	       (let ((l (read/rp *mpc-status-volume-grammar* ip)))
		  (cond
		     ((integer? l)
		      (loop l))
		     ((eq? l 'ok)
		      tm)
		     (else
		      (loop tm)))))))))

;*---------------------------------------------------------------------*/
;*    music-volume-set! ::mpc ...                                      */
;*---------------------------------------------------------------------*/
(define-method (music-volume-set! mpc::mpc vol)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-connect! mpc 'music-volume-set!)
	 (mpc-cmd mpc (string-append "setvol " (integer->string vol)))
	 (mpc-ok mpc))))

;*---------------------------------------------------------------------*/
;*    music-event-loop-continue ::mpc ...                              */
;*---------------------------------------------------------------------*/
(define-method (music-event-loop-continue mpc::mpc onstate onmeta onerror onvol)
   (let liip ()
      (with-handler
	 (lambda (exc)
	    (if (or (&io-timeout-error? exc) (&io-write-error? exc))
		(liip)
		(let loop ()
		   (unless (mpc-ok mpc)
		      (loop)))))))
   #t)
   
;*---------------------------------------------------------------------*/
;*    music-reset-error! ...                                           */
;*---------------------------------------------------------------------*/
(define-method (music-reset-error! mpc::mpc)
   (with-lock (mpc-%mutex mpc)
      (lambda ()
	 (mpc-cmd mpc "clearerror")
	 (mpc-ok mpc))))
   
