/*=====================================================================*/
/*    serrano/prgm/project/bigloo/runtime/Clib/cinit-obj.c             */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Tue Jan 29 09:19:48 2002                          */
/*    Last change :  Tue Feb 26 16:42:16 2008 (serrano)                */
/*    Copyright   :  2002-08 Manuel Serrano                            */
/*    -------------------------------------------------------------    */
/*    Bootstrap of pre-allocated objects.                              */
/*=====================================================================*/
#include <stdio.h>
#include <bigloo.h>

/*---------------------------------------------------------------------*/
/*    Les objects vitaux ...                                           */
/*---------------------------------------------------------------------*/
obj_t an_object;

obj_t quote;
obj_t *c_allocated_char;
BGL_RUNTIME_DEF obj_t _exit_value_;

obj_t bprof_port = BUNSPEC;

BGL_RUNTIME_DEF header_t bgl_opaque_nil = MAKE_HEADER( OPAQUE_TYPE, 0 );

/*---------------------------------------------------------------------*/
/*    Global mutex                                                     */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t bigloo_mutex = BUNSPEC;
BGL_RUNTIME_DEF obj_t bigloo_generic_mutex = BUNSPEC;
DEFINE_STRING( bigloo_mutex_name, _1, "bigloo-mutex", 12 );

/*---------------------------------------------------------------------*/
/*    Importations                                                     */
/*---------------------------------------------------------------------*/
extern obj_t string_to_obj();
extern void bgl_init_dynamic_env();
extern void bgl_init_symbol_table();
extern void bgl_init_socket();
extern void bgl_init_signal();
extern void bgl_init_keyword_table();
extern void bgl_init_io();
extern void bgl_init_trace();
extern void bgl_init_process_table();
extern void bgl_init_dload();

/*---------------------------------------------------------------------*/
/*    init_objects ...                                                 */
/*---------------------------------------------------------------------*/
void bgl_init_objects() {
   bgl_init_dynamic_env();
   bgl_init_trace();
   bgl_init_symbol_table();
   bgl_init_signal();
   bgl_init_io();
   bgl_init_keyword_table();
   bgl_init_process_table();
   bgl_init_dload();
   bgl_init_socket();

   bigloo_mutex = bgl_make_mutex( bigloo_mutex_name );
   bigloo_generic_mutex = bgl_make_mutex( bigloo_mutex_name );
   quote = string_to_symbol( "QUOTE" );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_module_debug_level ...                                       */
/*---------------------------------------------------------------------*/
static int bgl_module_debug_level = -1;

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    bgl_module_margins ...                                           */
/*---------------------------------------------------------------------*/
char *bgl_module_margins[] = { "",
			       " ",
			       "  ",
			       "   ",
			       "    ",
			       "     ",
			       "      ",
			       "       ",
			       "        ",
			       "         ",
			       "          ",
			       "           ",
			       "            ",
			       "             ",
			       "              ",
			       "               ",
			       "                " };

/*---------------------------------------------------------------------*/
/*    static char *                                                    */
/*    bgl_init_module_debug_margin ...                                 */
/*---------------------------------------------------------------------*/
static char *
margin( int level ) {
#define MAX_MODULE_DEBUG_LEVEL 16
   return bgl_module_margins[ bgl_module_debug_level < MAX_MODULE_DEBUG_LEVEL ?
			      bgl_module_debug_level : MAX_MODULE_DEBUG_LEVEL ];
}
   
/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_module_debug_start ...                                  */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_init_module_debug_start( char *mod ) {
   bgl_module_debug_level++;
   fprintf( stderr, "%s>>> %s (%d)\n",
	    margin( bgl_module_debug_level ),
	    mod,
	    bgl_module_debug_level );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_module_debug_library ...                                */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_init_module_debug_library( char *mod ) {
   fprintf( stderr, "%s library (%s)\n",
	    margin( bgl_module_debug_level ),
	    mod );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_module_debug_object ...                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_init_module_debug_object( char *mod ) {
   fprintf( stderr, "%s object (%s)\n",
	    margin( bgl_module_debug_level ),
	    mod );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_module_debug_import ...                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_init_module_debug_import( char *mod, char *import ) {
   fprintf( stderr, "%s import (%s) %s\n",
	    margin( bgl_module_debug_level ),
	    mod,
	    import );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bgl_init_module_debug_end ...                                    */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF void
bgl_init_module_debug_end( char *mod ) {
   fprintf( stderr, "%s<<< %s\n", margin( bgl_module_debug_level ), mod );
   bgl_module_debug_level--;
}
