/*=====================================================================*/
/*    serrano/prgm/project/bigloo/runtime/Clib/cports.c                */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Thu Jul 23 15:34:53 1992                          */
/*    Last change :  Mon May  5 11:20:39 2008 (serrano)                */
/*    -------------------------------------------------------------    */
/*    Input ports handling                                             */
/*=====================================================================*/
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _MSC_VER
#   include <dirent.h>
#   include <sys/time.h>
#else
#   include <io.h>
#   include <windows.h>
#   include <winsock2.h>
#   define pclose _pclose
#   define popen _popen
#   define S_ISDIR( mode ) ((mode & _S_IFDIR) != 0)
#endif
#include <string.h>
#if !defined( sony_news ) && \
    !(defined( NeXT ) && (defined( mc68000 ) || defined( i386 ))) && \
    !defined( _MSC_VER )
#   include <unistd.h>
#endif
#ifndef _MSC_VER
#  include <sys/file.h>
#endif
#include <bigloo.h>
#if( defined( sony_news ) || (defined( NeXT ) && defined( mc68000 )) )
#   include <ctype.h>
#endif

#if POSIX_FILE_OPS
#   include <unistd.h>
#endif

#if HAVE_TERMIOS
#  include <termios.h>
#else 
#  if HAVE_TERMIO
#    include <termio.h>
#  endif
#endif

/*---------------------------------------------------------------------*/
/*    DEBUG_SENDCHARS                                                  */
/*---------------------------------------------------------------------*/
#define DEBUG_SENDCHARS 1
#undef DEBUG_SENDCHARS

/*---------------------------------------------------------------------*/
/*    isascii                                                          */
/*---------------------------------------------------------------------*/
#if( !defined( isascii ) )
#   define isascii( c ) (!((c) & ~0177))
#endif

/*---------------------------------------------------------------------*/
/*    Global variables                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF long default_io_bufsiz;

/*---------------------------------------------------------------------*/
/*    External definitions.                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DECL obj_t make_input_port( char *, FILE *, obj_t, long );
extern obj_t make_string();
extern obj_t make_string_sans_fill();
extern long bgl_file_size( char * );

#if( BGL_HAVE_SENDFILE )
extern ssize_t sendfile( int, int, off_t *, size_t );
#endif

/*---------------------------------------------------------------------*/
/*    Prototypes                                                       */
/*---------------------------------------------------------------------*/
static bool_t pipe_name_p( char * );
static char *pipe_name(char *);

static size_t strwrite( void *, size_t, size_t, obj_t port );
static int strputc( int, obj_t );

static size_t procwrite( void *, size_t, size_t, obj_t port );
static int procputc( int, obj_t );

static size_t closedwrite( void *, size_t, size_t, obj_t port );
static int closedputc( int, obj_t );

/*---------------------------------------------------------------------*/
/*    struct sendfile_info_t ...                                       */
/*    -------------------------------------------------------------    */
/*    This structure is used for the Boehm's collection version        */
/*    7 and higher. The collector require a "blocking" function        */
/*    to be wrapping sendfile invokation. This function receives       */
/*    a sendfile_info_t as parameter and fills the "res" field.        */
/*---------------------------------------------------------------------*/
struct sendfile_info_t {
   int out;
   int in;
   off_t *offset;
   size_t sz;
   long res;
};

/*---------------------------------------------------------------------*/
/*    struct timeout_t ...                                             */
/*---------------------------------------------------------------------*/
struct timeout_t {
   long timeout;
   long (*sysread)();
   int (*sysputc)();
   size_t (*syswrite)();
};

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bglerror ...                                                     */
/*---------------------------------------------------------------------*/
static int
bglerror( int err ) {
   switch( err ) {
      case EPIPE: return BGL_IO_SIGPIPE_ERROR;

#if( defined( EBADF ) )
      case EBADF: return BGL_IO_PORT_ERROR;
#endif	 
#if( defined( EBADFD ) )
      case EBADFD: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENAMETOOLONG ) )
      case ENAMETOOLONG: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENFILE ) )
      case ENFILE: return BGL_IO_PORT_ERROR;
#endif
#if( defined( ENODEV ) )
      case ENODEV: return BGL_IO_PORT_ERROR;
#endif

#if( defined( ENOMEM ) )
      case ENOMEM: return BGL_IO_WRITE_ERROR;
#endif	 
#if( defined( ENOSPC ) )
      case ENOSPC: return BGL_IO_WRITE_ERROR;
#endif	 
	 
      default: return BGL_IO_ERROR;
   }
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_fputc ...                                                    */
/*---------------------------------------------------------------------*/
static int
bgl_fputc( int c, obj_t port ) {
   FILE *stream = PORT( port ).stream;
   
   return putc( c, stream );
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    bgl_fwrite ...                                                   */
/*---------------------------------------------------------------------*/
static size_t
bgl_fwrite( void *ptr, size_t size, size_t nmemb, obj_t port ) {
   return fwrite( ptr, size, nmemb, PORT( port ).stream );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    bgl_feof ...                                                     */
/*    -------------------------------------------------------------    */
/*    We explicitly define this function because on some system        */
/*    feof is a macro.                                                 */
/*---------------------------------------------------------------------*/
int
bgl_feof( obj_t obj ) {
   return feof( ((FILE *)PORT( obj ).stream) );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    proc_eof ...                                                     */
/*---------------------------------------------------------------------*/
static int
proc_eof( obj_t obj ) {
   return 1;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_fclose ...                                                   */
/*---------------------------------------------------------------------*/
static int
bgl_fclose( obj_t obj ) {
   return fclose( PORT( obj ).stream );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bgl_pclose ...                                                   */
/*---------------------------------------------------------------------*/
static int
bgl_pclose( obj_t obj ) {
   return pclose( PORT( obj ).stream );
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    plain_fread ...                                                  */
/*---------------------------------------------------------------------*/
static long
plain_fread( char *ptr, long size, long nmemb, obj_t port ) {
   FILE *stream = PORT( port ).stream;
   
#ifdef POSIX_FILE_OPS
   long num = size * nmemb;
   long n;

 loop:
   if( (n = read( fileno( stream ), ptr, num ) ) < 0 )
      if( errno == EINTR ) goto loop;

   return n;
#else
   long n;

 loop:
   if( !(n = fread( ptr, size, nmemb, stream )) ) {
      if( ferror( stream ) == EINTR ) {
	 clearerr( stream );
	 goto loop;
      } else {
	 C_SYSTEM_FAILURE( BGL_IO_READ_ERROR, "read", strerror(errno), port );
      }
   }

   return n;
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    console_fread ...                                                */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input, if */
/*    not enough characters are available.                             */
/*    -------------------------------------------------------------    */
/*    This function is prefered to bgl_pipe_fread for consoles because */
/*    it automatically flushes stdout before reading (see getc).       */
/*---------------------------------------------------------------------*/
static long
console_fread( char *ptr, long size, long nmemb, obj_t port ) {
   FILE *stream = PORT( port ).stream;
   long num = size * nmemb;
   char *buf = ptr;
   int c;

   while( ((c = getc( stream )) != EOF) ) {
      *buf++ = c;

      if( c == '\n' ) break;
      if( --num <= 0 ) break;
   }

   return (long)(buf - ptr);
}

/*---------------------------------------------------------------------*/
/*    long                                                             */
/*    bgl_pipe_fread ...                                               */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input if  */
/*    insufficient characters are available.                           */
/*---------------------------------------------------------------------*/
long
bgl_pipe_fread( char *ptr, long size, long nmemb, obj_t port ) {
#ifdef POSIX_FILE_OPS
   FILE *stream = PORT( port ).stream;
   int fd = fileno( stream );
   
   long num = size * nmemb;
   long n;

 loop:
   if( (n = read( fd, ptr, num )) < 0 ) {
      if( errno == EINTR ) goto loop;
      C_SYSTEM_FAILURE( bglerror( errno ), "read", strerror( errno ), port );
   }

#if( BGL_HAVE_SELECT )
   if( n == 0 ) {
      /* is it a true end-of-file or a simple lack of input? */
      struct timeval timeout;
      fd_set readfds;
      
      timeout.tv_sec = 0;
      timeout.tv_usec = 10000;

      FD_ZERO( &readfds );
      FD_SET( fileno( stream ), &readfds );

      if( (n = select( fd + 1, &readfds, NULL, NULL, &timeout ) <= 0 ) ) {
	 if( n == 0 ) {
	    /* FIXME */
	    fprintf( stderr, "%s:%d:LOOPING on select timeout\n", __FILE__, __LINE__ );
	    goto loop;
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			      "read/timeout",
			      strerror( errno ),
			      port );
	 }
      }
   }
#endif
   return n;
#else
   return console_fread( ptr, size, nmemb, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    proc_read ...                                                    */
/*---------------------------------------------------------------------*/
static long
proc_read( char *b, long s, long n, obj_t port ) {
   long l = s * n;
   obj_t buf = CREF(port)->input_port_t.pbuffer;

 loop:

   if( STRINGP( buf ) ) {
      /* won't read because the proc buffer is already filled */
      char *s = BSTRING_TO_STRING( buf );
      long p = CREF( port )->input_port_t.pbufpos;
      long r = STRING_LENGTH( buf ) - p;

      if( r <= l ) {
	 memmove( b, &s[ p ], r );
	 CREF( port )->input_port_t.pbuffer = BFALSE;
	 CREF( port )->input_port_t.pbufpos = 0;
	 return r;
      } else {
	 memmove( b, &s[ p ], l );
	 CREF( port )->input_port_t.pbufpos += l;
	 return l;
      }
   } else {
      /* invoke the procedure to fill the buffer */
      obj_t proc = CREF( port )->input_port_t.port.name;
      obj_t nbuf = PROCEDURE_ENTRY( proc )( proc, BEOA );

      if( STRINGP( nbuf ) ) {
	 buf = CREF( port )->input_port_t.pbuffer = nbuf;
	 goto loop;
      } else {
	 if( nbuf == BFALSE ) {
	    /* eof has been reached */
	    CREF( port )->input_port_t.eof = 1;
	    return 0;
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "input-procedure-port",
			      "Procedure result must be #f, or a string",
			      nbuf);
	 }
      }
   }
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    timeout_set_port_blocking ...                                    */
/*---------------------------------------------------------------------*/
#if( BGL_HAVE_FCNTL )
static void
timeout_set_port_blocking( char *fun, int fd, int bool ) {
   int val;

   if( (val = fcntl( fd, F_GETFL, 0 ) ) < 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_ERROR, fun, strerror( errno ), BINT( fd ) );
   }

   if( !bool ) 
      val |= O_NONBLOCK;
   else 
      val &= ~O_NONBLOCK;

   if( fcntl( fd, F_SETFL, val ) < 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_ERROR, fun, strerror( errno ), BINT( fd ) );
   }
}
#endif 

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    sysputc_with_timeout ...                                         */
/*---------------------------------------------------------------------*/
static int
sysputc_with_timeout( int c, obj_t port ) {
#if( defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   FILE *stream = PORT( port ).stream;
   struct timeout_t *tmt = PORT( port ).timeout;
   int (*sysputc)() = tmt->sysputc;
   int n;
   
   long to = tmt->timeout;

   fd_set writefds;
   struct timeval timeout;
   int fd = fileno( stream );

   FD_ZERO( &writefds );
   FD_SET( fd, &writefds );

   timeout.tv_sec = to / 1000000;
   timeout.tv_usec = to % 1000000;

   if( (n = select( fd + 1, NULL, &writefds, NULL, &timeout )) <= 0 ) {
      if( n == 0 ) {
	 C_SYSTEM_FAILURE( BGL_IO_TIMEOUT_ERROR,
			   "putc/timeout",
			   "time limit exceeded",
			   port );
      } else {
	 C_SYSTEM_FAILURE( BGL_IO_WRITE_ERROR,
			   "putc/timeout",
			   strerror( errno ),
			   port );
	 
      }
   } else {
      return sysputc( c, port );
   }
#else
   return BGL_DISPLAY_CHAR( c, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    syswrite_with_timeout ...                                        */
/*---------------------------------------------------------------------*/
static size_t
syswrite_with_timeout( void *ptr, size_t size, size_t nmemb, obj_t port ) {
#if( defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   FILE *stream = PORT( port ).stream;
   struct timeout_t *tmt = PORT( port ).timeout;
   size_t (*syswrite)() = tmt->syswrite;
   
   long to = tmt->timeout;
   long count = 0;
   long sum = nmemb * size;

   fd_set writefds;
   struct timeval timeout;
   int fd = fileno( stream );

   FD_ZERO( &writefds );
   FD_SET( fd, &writefds );

   timeout.tv_sec = to / 1000000;
   timeout.tv_usec = to % 1000000;

   while( count < sum ) {
      int n;
      
      if( n = select( fd + 1, NULL, &writefds, NULL, &timeout ) <= 0 ) {
	 if( n == 0 ) {
	    C_SYSTEM_FAILURE( BGL_IO_TIMEOUT_ERROR,
			      "write/timeout",
			      "time limit exceeded",
			      port );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_WRITE_ERROR,
			      "write/timeout",
			      strerror( errno ),
			      port );
	 }
      } else {
	 int n = syswrite( &((char *) ptr)[ count ], 1, sum - count, port );

	 if (n >= 0) {
	    count += n;
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_WRITE_ERROR,
			      "write/timeout",
			      strerror( errno ),
			      port );
	 }
      }
   }

   return count;
#else
   return OUTPUT_PORT( port ).syswrite( ptr, size, nmemb, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    sysread_with_timeout ...                                         */
/*    -------------------------------------------------------------    */
/*    In constrast to fread, this function does not block on input if  */
/*    insufficient characters are available.                           */
/*---------------------------------------------------------------------*/
static long
sysread_with_timeout( char *ptr, long size, long nmemb, obj_t port ) {
#if( defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL )
   FILE *stream = PORT( port ).stream;
   struct timeout_t *tmt = PORT( port ).timeout;
   long (*sysread)() = tmt->sysread;
   long n;

   long to = tmt->timeout;

   fd_set readfds;
   struct timeval timeout;
   int fd = fileno( stream );

   FD_ZERO( &readfds );
   FD_SET( fd, &readfds );

   timeout.tv_sec = to / 1000000;
   timeout.tv_usec = to % 1000000;

loop:   
   if( (n = select( fd + 1, &readfds, NULL, NULL, &timeout )) <= 0 ) {
      if( n == 0 ) {
	 C_SYSTEM_FAILURE( BGL_IO_TIMEOUT_ERROR,
			   "read/timeout",
			   "time limit exceeded",
			   port );
      } else {
	 /* MS: 23 Jan 2008. No attention was paid to EINTR */
	 /* I'm not 100% sure that this new test is correct */
	 if( errno == EINTR ) {
	    fprintf( stderr, "*** DEBUG SELECT receives EINTR: %s:%d\n",
		     __FILE__, __LINE__ );
	    FD_ZERO( &readfds );
	    FD_SET( fd, &readfds );

	    timeout.tv_sec = to / 1000000;
	    timeout.tv_usec = to % 1000000;
	    goto loop;
	 }
	 
	 C_SYSTEM_FAILURE( BGL_IO_READ_ERROR,
			   "read/timeout",
			   strerror( errno ),
			   port );
      }
   } else {
      n = sysread( ptr, size, nmemb, port );

      return n;
   }
#else
   return INPUT_PORT( port ).sysread( ptr, size, nmemb, port );
#endif
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    fileflush ...                                                    */
/*---------------------------------------------------------------------*/
static obj_t
fileflush( obj_t o ) {
   return ((fflush( PORT( o ).stream ) == EOF) ? BFALSE : BTRUE);
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    closedflush ...                                                  */
/*---------------------------------------------------------------------*/
static obj_t
closedflush( obj_t o ) {
   ;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    strflush ...                                                     */
/*---------------------------------------------------------------------*/
static obj_t
strflush( obj_t port ) {
   obj_t res;
   
   if( OUTPUT_STRING_PORT( port ).buffer ) {
      return string_to_bstring_len( OUTPUT_STRING_PORT( port ).buffer,
				    OUTPUT_STRING_PORT( port ).offset );
   } else {
      return string_to_bstring( "" );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    procflush ...                                                    */
/*---------------------------------------------------------------------*/
static obj_t
procflush( obj_t port ) {
   obj_t flush = OUTPUT_PROCEDURE_PORT( port ).flush;

   return PROCEDURE_ENTRY( flush )( flush, BEOA );
}

/*---------------------------------------------------------------------*/
/*     make_output_port ...                                            */
/*    -------------------------------------------------------------    */
/*    Since bigloo3.0a (9 may 2007), sysputc defaults to 0. Hence,     */
/*    each use of sysputc (mainly BGL_DISPLAY_CHAR) must check         */
/*    it prior to calling the function. When sysputc is 0, the         */
/*    C macro PUTC must be called instead.                             */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
make_output_port( char *name, FILE * file, obj_t kindof ) {
   obj_t new_output_port;

   new_output_port = GC_MALLOC( OUTPUT_PORT_SIZE );

   new_output_port->output_port_t.port.header = MAKE_HEADER( OUTPUT_PORT_TYPE, 0 );
   new_output_port->port_t.stream = file;
   new_output_port->port_t.name = string_to_bstring( name );
   new_output_port->port_t.kindof = kindof;
   new_output_port->port_t.sysseek = 0;
   new_output_port->port_t.chook = BUNSPEC;
   new_output_port->port_t.userdata = BUNSPEC;
   new_output_port->port_t.timeout = 0L;
   new_output_port->output_port_t.sysputc = bgl_fputc; 
   new_output_port->output_port_t.syswrite = bgl_fwrite;
   new_output_port->output_port_t.sysflush = fileflush;

   switch( (long) kindof ) {
      case (long) KINDOF_FILE:
      case (long) KINDOF_PROCPIPE:
	 new_output_port->port_t.sysclose = bgl_fclose;
	 break;

      case (long) KINDOF_PIPE:
	 new_output_port->port_t.sysclose = bgl_pclose;
	 break;

      case (long) KINDOF_CONSOLE:
	 new_output_port->port_t.sysclose = 0L;
	 break;

      default:
	 new_output_port->port_t.sysclose = 0L;
   }

   return BREF( new_output_port );
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    bgl_output_port_timeout_set ...                                  */
/*---------------------------------------------------------------------*/
bool_t
bgl_output_port_timeout_set( obj_t port, long timeout ) {
#if defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL
   if( (PORT( port ).kindof == KINDOF_FILE) ||
       (PORT( port ).kindof == KINDOF_PIPE) ||
       (PORT( port ).kindof == KINDOF_PROCPIPE) ||
       (PORT( port ).kindof == KINDOF_CONSOLE) ||
       (PORT( port ).kindof == KINDOF_SOCKET) ) {

      if( timeout == 0 ) {
	 struct timeout_t *to = PORT( port ).timeout;

	 if( to ) {
	    OUTPUT_PORT( port ).syswrite = to->syswrite;
	    OUTPUT_PORT( port ).sysputc = to->sysputc;
	    PORT( port ).timeout = 0L;

	    timeout_set_port_blocking( "output-port-timeout-set!",
				       fileno( (FILE *)(PORT( port ).stream) ),
				       1 );
	 }

	 return 0;
      } else {
	 struct timeout_t *to =
	    (struct timeout_t *)GC_MALLOC( sizeof( struct timeout_t ) );

	 to->timeout = timeout;

	 if( PORT( port ).timeout ) {
	    to->syswrite =
	       ((struct timeout_t *)PORT( port ).timeout)->syswrite;
	    to->sysputc =
	       ((struct timeout_t *)PORT( port ).timeout)->sysputc;
	    
	    PORT( port ).timeout = (FILE *)to;
	 } else {
	    to->syswrite = OUTPUT_PORT( port ).syswrite;
	    to->sysputc = OUTPUT_PORT( port ).sysputc;
	    
	    OUTPUT_PORT( port ).syswrite = &syswrite_with_timeout;
	    OUTPUT_PORT( port ).sysputc = &sysputc_with_timeout;
	    
	    PORT( port ).timeout = (FILE *)to;
	    
	    timeout_set_port_blocking( "output-port-timeout-set!",
				       fileno( (FILE *)(PORT( port ).stream) ),
				       0 );

	 }
	 
	 return 1;
      }
   }
#endif

   return 0;
}

/*---------------------------------------------------------------------*/
/*    open_output_file ...                                             */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_output_file( obj_t name ) {
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      if( !(file = popen( pipe_name( cname ), "w" )) )
         return BFALSE;

       return make_output_port( cname, file, KINDOF_PIPE );
   } else
#endif
   {
      if (strcmp( cname, "null:" ) == 0)
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif

      if( !(file = fopen( cname, "wb" )) )
         return BFALSE;

      return make_output_port( cname, file, KINDOF_FILE );
   }
}

/*---------------------------------------------------------------------*/
/*    append_output_file ...                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
append_output_file( obj_t name ) {
   FILE *file;

   if( !(file = fopen( BSTRING_TO_STRING(name), "a+b ")) )
      return BFALSE;

   return make_output_port( BSTRING_TO_STRING( name ),
			    file,
			    KINDOF_FILE );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_output_string ...                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_output_string() {
   char *buffer;
   obj_t port;

   port = GC_MALLOC( OUTPUT_STRING_PORT_SIZE );

   /* We allocate a buffer filled of zero */
   buffer = (char *)(GC_MALLOC_ATOMIC( OUTPUT_STRING_PORT_BUFFER_SIZE + 1 ));
   bzero( buffer, OUTPUT_STRING_PORT_BUFFER_SIZE + 1 );
   
   port->port_t.header = MAKE_HEADER( OUTPUT_STRING_PORT_TYPE, 0 );
   port->port_t.kindof = KINDOF_STRING;
   port->port_t.sysclose = 0L;
   port->port_t.sysseek = 0L;
   port->port_t.chook = BUNSPEC;
   port->port_t.userdata = BUNSPEC;
   port->port_t.stream = port;
   port->output_port_t.sysputc = strputc;
   port->output_port_t.syswrite = strwrite;
   port->output_port_t.sysflush = strflush;
   
   port->output_string_port_t.buffer = buffer;
   port->output_string_port_t.size = OUTPUT_STRING_PORT_BUFFER_SIZE;
   port->output_string_port_t.offset = 0;

   return BREF( port );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_output_procedure ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_output_procedure( obj_t proc, obj_t flush ) {
   obj_t port;

   port = GC_MALLOC( OUTPUT_PROCEDURE_PORT_SIZE );

   port->port_t.header = MAKE_HEADER( OUTPUT_PROCEDURE_PORT_TYPE, 0 );
   port->port_t.kindof = KINDOF_PROCEDURE;
   port->port_t.sysclose = 0L;
   port->port_t.sysseek = 0L;
   port->port_t.chook = BUNSPEC;
   port->port_t.userdata = BUNSPEC;
   port->port_t.stream = port;
   port->output_port_t.sysputc = procputc;
   port->output_port_t.syswrite = procwrite;
   port->output_port_t.sysflush = procflush;
   
   port->output_procedure_port_t.buflen = 10;
   port->output_procedure_port_t.buffer = make_string( 80 );
   port->output_procedure_port_t.proc = proc;
   port->output_procedure_port_t.flush = flush;

   return BREF( port );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    get_output_string ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
get_output_string( obj_t port ) {
   if( OUTPUT_STRING_PORTP( port ) ) {
      if( OUTPUT_STRING_PORT (port ).size == 0 )
	 return make_string( 0, ' ' );
      else
	 return string_to_bstring_len( OUTPUT_STRING_PORT( port ).buffer,
				       OUTPUT_STRING_PORT( port ).offset);
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"get-output-string",
			"Not a string port",
			port );
      return BUNSPEC;
   }
}

/*---------------------------------------------------------------------*/
/*    close_output_port ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
close_output_port( obj_t port ) {
   obj_t res = port;

   if( OUTPUT_PORTP( port ) && (PORT( port ).kindof != KINDOF_CLOSED) ) {
      obj_t chook = PORT_CHOOK( port );
      
      PORT( port ).kindof = KINDOF_CLOSED;
      
      if( OUTPUT_STRING_PORTP( port ) ) {
	 res = get_output_string( port );
	 GC_free( (obj_t)(CREF( port )->output_string_port_t.buffer) );
	 CREF( port )->output_string_port_t.buffer = 0;
      }

      OUTPUT_PORT( port ).sysputc = closedputc;
      OUTPUT_PORT( port ).syswrite = closedwrite;
      
      if( PORT( port ).sysclose ) PORT( port ).sysclose( port );

      OUTPUT_PORT( port ).sysflush = closedflush;

      PORT( port ).sysclose = 0L;
      
      if( PROCEDUREP( chook ) ) {
	 if( PROCEDURE_ARITY( chook ) == 1 ) {
	    PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	 } else {
	    C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			      "close-output-port",
			      "illegal close hook arity",
			      chook );
	 }
      }
   }
   
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    make_input_port ...                                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
make_input_port( char *name, FILE *file, obj_t kindof, long bufsiz ) {
   obj_t new_input_port;

   /* An input port cannot be allocated as an atomic object    */
   /* because it holds a buffer and a name that are GC objects */
   new_input_port = GC_MALLOC( INPUT_PORT_SIZE );

   new_input_port->port_t.header = MAKE_HEADER( INPUT_PORT_TYPE, 0 );
   new_input_port->port_t.kindof = kindof;
   new_input_port->port_t.name = string_to_bstring( name );
   new_input_port->port_t.stream = file;
   new_input_port->port_t.timeout = 0L;
   new_input_port->port_t.sysseek = 0L;
   new_input_port->port_t.chook = BUNSPEC;
   new_input_port->port_t.userdata = BUNSPEC;
   new_input_port->input_port_t.filepos = 0;
   new_input_port->input_port_t.fillbarrier = -1;
   new_input_port->input_port_t.bufsiz = bufsiz;
   new_input_port->input_port_t.eof = 0;
   new_input_port->input_port_t.matchstart = 0;
   new_input_port->input_port_t.matchstop = 0;
   new_input_port->input_port_t.forward = 0;
   new_input_port->input_port_t.bufpos = 1;
   new_input_port->input_port_t.lastchar = '\n';
   new_input_port->input_port_t.syseof = bgl_feof;
   new_input_port->input_port_t.pbuffer = BFALSE;
   new_input_port->input_port_t.pbufpos = 0;
   new_input_port->input_port_t.gzip = BFALSE;

   switch( (long) kindof ) {
      case (long)KINDOF_CONSOLE:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.sysread = console_fread;
	 break;

#if( HAVE_PIPE )
      case (long)KINDOF_PIPE:
	 new_input_port->port_t.sysclose = bgl_pclose;
	 new_input_port->input_port_t.sysread = bgl_pipe_fread;
	 break;

      case (long)KINDOF_PROCPIPE:
#endif
      case (long)KINDOF_SOCKET:
	 new_input_port->port_t.sysclose = bgl_fclose;
	 new_input_port->input_port_t.sysread = bgl_pipe_fread;
	 break;

      case (long)KINDOF_FILE:
	 new_input_port->port_t.sysclose = bgl_fclose;
	 new_input_port->input_port_t.sysread = plain_fread;
	 break;

      case (long)KINDOF_PROCEDURE:
      case (long)KINDOF_GZIP:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.syseof = proc_eof;
	 new_input_port->input_port_t.sysread = proc_read;
	 break;

      default:
	 new_input_port->port_t.sysclose = 0;
	 new_input_port->input_port_t.sysread = plain_fread;
   }


   if( bufsiz > 0 ) {
      new_input_port->input_port_t.buffer =
	 (unsigned char *) GC_MALLOC_ATOMIC( bufsiz + 1 );
      new_input_port->input_port_t.buffer[ 0 ] = '\0';
   } else {
      new_input_port->input_port_t.buffer = '\0';
   }

   return BREF( new_input_port );
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    bgl_input_port_timeout_set ...                                   */
/*---------------------------------------------------------------------*/
bool_t
bgl_input_port_timeout_set( obj_t port, long timeout ) {
#if defined( POSIX_FILE_OPS ) && BGL_HAVE_SELECT && BGL_HAVE_FCNTL
   if( (PORT(port).kindof == KINDOF_FILE) ||
       (PORT(port).kindof == KINDOF_PIPE) ||
       (PORT(port).kindof == KINDOF_PROCPIPE) ||
       (PORT(port).kindof == KINDOF_CONSOLE) ||
       (PORT(port).kindof == KINDOF_SOCKET) ) {
      if( timeout == 0 ) {
	 struct timeout_t *to = PORT( port ).timeout;

	 if( to ) {
	    INPUT_PORT( port ).sysread = to->sysread;
	    PORT( port ).timeout = 0L;

	    timeout_set_port_blocking( "input-port-timeout-set!",
				       fileno( (FILE *)(PORT( port ).stream) ),
				       1 );
	 }

	 return 0;
      } else {
	 struct timeout_t *to =
	    (struct timeout_t *)GC_MALLOC( sizeof( struct timeout_t ) );

	 to->timeout = timeout;
	 
	 if( PORT( port ).timeout ) {
	    to->sysread =
	       ((struct timeout_t *)PORT( port ).timeout)->sysread; 
	    
	    PORT( port ).timeout = (FILE *)to;
	 } else {
	    to->sysread = INPUT_PORT( port ).sysread;

	    if( fileno( (FILE *)(PORT( port ).stream) ) == -1 ) {
	       C_SYSTEM_FAILURE( bglerror( errno ),
				 "input-port-timeout-set!",
				 "Illegal input-port",
				 port );
	    }

	    INPUT_PORT( port ).sysread = &sysread_with_timeout;

	    PORT( port ).timeout = (FILE *)to;
	    
	    timeout_set_port_blocking( "input-port-timeout-set!",
				       fileno( (FILE *)(PORT( port ).stream) ),
				       0);
	 }

	 return 1;
      }
   }
#endif

   return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_input_pipe ...                                              */
/*---------------------------------------------------------------------*/
obj_t
open_input_pipe( obj_t name, obj_t bbufsiz ) {
#if( HAVE_PIPE )
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

   if( !(file = popen( cname, "r" )) )
      return BFALSE;

   /* we use our own buffer */
   setvbuf( file, NULL, _IONBF, 0 );  

   return make_input_port( cname, file, KINDOF_PIPE, CINT( bbufsiz ) );
#else
   return BFLASE;
#endif
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_input_resource ...                                          */
/*    -------------------------------------------------------------    */
/*    Resources are only meaning in Java.                              */
/*---------------------------------------------------------------------*/
obj_t
open_input_resource( obj_t name, obj_t bbufsiz ) {
   return BFALSE;
}

/*---------------------------------------------------------------------*/
/*    open_input_file ...                                              */
/*    -------------------------------------------------------------    */
/*    !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!! WARNING !!!  */
/*    We fill up its associated buffer when opening an input port.     */
/*    -------------------------------------------------------------    */
/*    This function open two kind of files. Regular file and Unix      */
/*    like pipes when the file name is something like "| ...".         */
/*---------------------------------------------------------------------*/
obj_t
open_input_file( obj_t name, obj_t bbufsiz ) {
   FILE *file;
   char *cname = BSTRING_TO_STRING( name );

#if( HAVE_PIPE )
   if( pipe_name_p( cname ) ) {
      if( !(file = popen( pipe_name( cname ), "r" ) ) )
	 return BFALSE;

      /* we use our own buffer */
      setvbuf( file, NULL, _IONBF, 0 );  

      return make_input_port( cname, file, KINDOF_PIPE, CINT( bbufsiz ) );
   } else
#endif
   {
      if (strcmp( cname, "null:" ) == 0)
#        ifndef _MSC_VER
            cname= "/dev/null";
#        else
            cname= "NUL:";
#        endif

       if( !(file = fopen( cname, "rb" )) ) {
	  return BFALSE;
       } else {
	  long sz = bgl_file_size( cname );
	  long bufsiz = CINT( bbufsiz );

	  switch( sz ) {
	     case 0:
		break;
	     case 1:
		bufsiz = 3;
		break;

	     default:
		if( (sz < bufsiz) ) bufsiz = (long)sz + 1;
	  }

	  /* we use our own buffer */
	  setvbuf( file, NULL, _IONBF, 0 );  

	  return make_input_port( cname, file, KINDOF_FILE, bufsiz );
       }
   }
}

/*---------------------------------------------------------------------*/
/*    open_input_console ...                                           */
/*---------------------------------------------------------------------*/
obj_t
open_input_console() {
   return make_input_port("[stdin]", stdin, KINDOF_CONSOLE, default_io_bufsiz);
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    file_to_buffered_input_port ...                                  */
/*---------------------------------------------------------------------*/
obj_t
file_to_buffered_input_port( FILE *file, long bufsize ) {
   if( file == stdin )
      return open_input_console();
   else {
      return make_input_port( "[file]", file, KINDOF_FILE, bufsize );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    file_to_input_port ...                                           */
/*---------------------------------------------------------------------*/
obj_t
file_to_input_port( FILE * file ) {
   return file_to_buffered_input_port( file, -1 );
}

/*---------------------------------------------------------------------*/
/*    open_input_string ...                                            */
/*    -------------------------------------------------------------    */
/*    The start index is in bounds.                                    */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_open_input_string( obj_t string, int start ) {
   obj_t port;
   long bufsiz = STRING_LENGTH( string ) - start;

   port = make_input_port( "[string]", 0L, KINDOF_STRING, bufsiz + 1 );

   CREF(port)->input_port_t.eof = 1;
   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   memcpy( &RGC_BUFFER(port)[ 0 ], &(STRING_REF( string, start )), bufsiz );
   RGC_BUFFER(port)[ bufsiz ] = '\0';

   return port;
}

/*---------------------------------------------------------------------*/
/*    open_input_procedure ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_input_procedure( obj_t fun, obj_t bfz ) {
   if( PROCEDURE_CORRECT_ARITYP( fun, 0 ) ) {
      obj_t port = make_input_port( "", 0L, KINDOF_PROCEDURE, CINT( bfz )+1 );

      CREF( port )->port_t.stream = port;
      CREF( port )->port_t.name = fun;
      CREF( port )->input_port_t.pbuffer = BUNSPEC;
      CREF( port )->input_port_t.pbufpos = 0;

      return port;
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"open-input-procedure",
			"Illegal procedure arity",
			fun );
   }
}

/*---------------------------------------------------------------------*/
/*    open_input_gzip_port ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_input_gzip_port( obj_t fun, obj_t in, obj_t bfz ) {
   if( PROCEDURE_CORRECT_ARITYP( fun, 0 ) ) {
      obj_t port = make_input_port( "", 0L, KINDOF_GZIP, CINT( bfz )+1 );

      CREF( port )->port_t.stream = port;
      CREF( port )->port_t.name = fun;
      CREF( port )->input_port_t.pbuffer = BUNSPEC;
      CREF( port )->input_port_t.pbufpos = 0;
      CREF( port )->input_port_t.gzip = in;

      return port;
   } else {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
			"open-input-gzip-port",
			"Illegal procedure arity",
			fun );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    open_input_c_string ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
open_input_c_string( char *c_string ) {
   obj_t port;
   long bufsiz = (long)strlen( c_string );
   char *new_string = 0;

   port = make_input_port( "[c_string]", 0L, KINDOF_STRING, 0 );

   CREF(port)->input_port_t.eof = 1;
   CREF(port)->input_port_t.bufsiz = bufsiz + 1;
   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   CREF(port)->input_port_t.buffer = (unsigned char *)c_string;

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    reopen_input_c_string ...                                        */
/*    -------------------------------------------------------------    */
/*    Simply changes the input buffer of an input string. Does not     */
/*    re-allocate a brand new input-port.                              */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
reopen_input_c_string( obj_t port, char *c_string ) {
   long bufsiz = (long)strlen(c_string);

   if( CREF(port)->input_port_t.bufsiz < (bufsiz + 1) ) {
      CREF(port)->input_port_t.bufsiz = bufsiz + 1;
      CREF(port)->input_port_t.buffer =
	 (unsigned char *)GC_MALLOC_ATOMIC(bufsiz + 1);
   }

   CREF(port)->input_port_t.bufpos = bufsiz + 1;
   CREF(port)->input_port_t.matchstart = 0;
   CREF(port)->input_port_t.matchstop = 0;
   CREF(port)->input_port_t.forward = 0;
   CREF(port)->input_port_t.lastchar = '\n';
   strcpy( (char *)(CREF(port)->input_port_t.buffer), (char *)c_string );

   return port;
}

/*---------------------------------------------------------------------*/
/*    close_input_port ...                                             */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
close_input_port( obj_t port ) {
   if( INPUT_PORTP( port ) ) {
      if( (PORT( port ).kindof != KINDOF_CLOSED) &&
	  (PORT( port ).kindof != KINDOF_CONSOLE) ) {
	 obj_t chook = PORT_CHOOK( port );
	 
	 if( PORT( port ).sysclose ) PORT( port ).sysclose( port );
	 
	 /* added by MS on 11oct2005 */
	 if( (PORT( port ).kindof != KINDOF_STRING) )
	    GC_free( (obj_t)(INPUT_PORT( port ).buffer) );
	 INPUT_PORT( port ).buffer = 0L;
	 
	 INPUT_PORT( port ).eof = 1;
      	 PORT( port ).kindof = KINDOF_CLOSED;
	 PORT( port ).sysclose = 0L;
	 
	 if( PROCEDUREP( chook ) ) {
	    if( PROCEDURE_ARITY( chook ) == 1 ) {
	       PROCEDURE_ENTRY( chook )( chook, port, BEOA );
	    } else {
	       C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR,
				 "close-input-port",
				 "illegal close hook arity",
				 chook );
	    }
	 }
      }

   }

   return port;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_seek ...                                          */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_input_port_seek( obj_t port, long pos ) {
   if( INPUT_PORT_ON_FILEP( port ) ) {
      if( fseek( PORT( port ).stream, pos, SEEK_SET ) )
	 return BFALSE;
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).eof = 0;
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).forward = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).lastchar = '\n';
      INPUT_PORT( port ).buffer[0] = '\0';

      return BTRUE;
   }

   if( INPUT_PORT_ON_STRINGP( port ) && (pos < INPUT_PORT( port ).bufsiz) ) {
      INPUT_PORT( port ).filepos = pos;
      INPUT_PORT( port ).matchstart = pos;
      INPUT_PORT( port ).matchstop = pos;
      INPUT_PORT( port ).forward = pos;

      return BTRUE;
   }

   return BFALSE;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_output_port_seek ...                                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_output_port_seek( obj_t port, long pos ) {
   if( OUTPUT_FILE_PORTP( port ) ) {
      if( fseek( PORT( port ).stream, pos, SEEK_SET ) )
	 return BFALSE;
      return BTRUE;
   }
   if( OUTPUT_STRING_PORTP( port ) ) {
      if( pos >= OUTPUT_STRING_PORT( port ).size )
	 return BFALSE;

      OUTPUT_STRING_PORT( port ).offset = pos;
      return BTRUE;
   }
   return BFALSE;
}
   
/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_input_port_reopen ...                                        */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_input_port_reopen( obj_t port ) {
   FILE *nf;

   if( !INPUT_PORT_ON_FILEP( port ) ) {
      if( INPUT_STRING_PORTP( port ) ) {
 	 return bgl_input_port_seek( port, 0 );
      } else {
	 return BFALSE;
      }
   }
   
   nf = freopen( BSTRING_TO_STRING( PORT( port ).name ),
		 "r",
		 (FILE *)PORT( port ).stream );

   if( !nf ) return BFALSE;
   
   PORT( port ).stream = (void *)nf;
      
   /* we use our own buffer */
   setvbuf( (FILE *)PORT( port ).stream, NULL, _IONBF, 0 );

   INPUT_PORT( port ).filepos = 0;
   INPUT_PORT( port ).eof = 0;
   INPUT_PORT( port ).matchstart = 0;
   INPUT_PORT( port ).matchstop = 0;
   INPUT_PORT( port ).forward = 0;
   INPUT_PORT( port ).bufpos = 1;
   INPUT_PORT( port ).lastchar = '\n';
   INPUT_PORT( port ).buffer[ 0 ] = '\0';

   return BTRUE;
}

/*---------------------------------------------------------------------*/
/*    obj                                                              */
/*    reset_console ...                                                */
/*    -------------------------------------------------------------    */
/*    We flush input port, for ^C to work correctly within the         */
/*    interpreter. The only place this function is called is in the    */
/*    REPL (see Eval/eval.scm).                                        */
/*---------------------------------------------------------------------*/
obj_t
reset_console( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_CONSOLE ) {
      INPUT_PORT( port ).matchstart = 0;
      INPUT_PORT( port ).matchstop = 0;
      INPUT_PORT( port ).bufpos = 1;
      INPUT_PORT( port ).buffer[ 0 ] = '\0';
      INPUT_PORT( port ).lastchar = '\n';
   }

   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*     bgl_init_io ...                                                 */
/*---------------------------------------------------------------------*/
void
bgl_init_io() {
#if( !defined( _SBFSIZ ) )
#   define _SBFSIZ 1
#endif
   obj_t denv = BGL_CURRENT_DYNAMIC_ENV();

   default_io_bufsiz = BUFSIZ * _SBFSIZ;

   BGL_ENV_CURRENT_OUTPUT_PORT_SET(
      denv,
      make_output_port( "stdout", stdout, KINDOF_CONSOLE ) );
   BGL_ENV_CURRENT_ERROR_PORT_SET(
      denv,
      make_output_port( "stderr", stderr, KINDOF_CONSOLE ) );
   BGL_ENV_CURRENT_INPUT_PORT_SET(
      denv,
      open_input_console() );
}

/*---------------------------------------------------------------------*/
/*    fexists ...                                                      */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
fexists( char *name ) {
#if( HAVE_PIPE )
   if( pipe_name_p( name ) ) {
      return 1;
   }
#else
   if( pipe_name_p(name) ) {
      return 0;
   }
#endif

# ifndef _MSC_VER
    return !access( name, F_OK );
# else
    /* !!!!! verify semantics of Unix' access */
    return !_access( name, 0 );        
# endif
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    reset_eof ...                                                    */
/*    -------------------------------------------------------------    */
/*    The function erase the end-of-file of input console port.        */
/*    This allows, restart reading after a ^D.                         */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
reset_eof( obj_t port ) {
   if( PORT( port ).kindof == KINDOF_CONSOLE ) {
      /* we forget about EOF */
      INPUT_PORT( port ).eof = 0;

      /* we cleanup buffer   */
      reset_console( port );

      /* we clear errors.    */
      clearerr( stdin );

      return 1;
   }
   else
      return 0;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_strport_grow ...                                             */
/*---------------------------------------------------------------------*/
obj_t
bgl_strport_grow( obj_t p ) {
   char *buffer = OUTPUT_STRING_PORT(p).buffer;
   long old_size = OUTPUT_STRING_PORT(p).size;
   long new_size;

   if( old_size == 0 ) {
      C_SYSTEM_FAILURE( BGL_IO_PORT_ERROR, "string-port", "port closed", p );

      return p;
   } else {
      new_size = old_size * 2;
      buffer = (char *)GC_realloc( (obj_t)buffer, new_size + 1 );
      buffer[ old_size ] ='\0';
      OUTPUT_STRING_PORT(p).buffer = buffer;

      OUTPUT_STRING_PORT(p).size = new_size;

      return p;
   }
}

/*---------------------------------------------------------------------*/
/*    static bool_t                                                    */
/*    pipe_name_p ...                                                  */
/*    -------------------------------------------------------------    */
/*    Is a file name a pipe name ? A pipe name start by the            */
/*    sequence "| ".                                                   */
/*---------------------------------------------------------------------*/
static bool_t
pipe_name_p( char *name ) {
   return( (name[ 0 ] == '|') && (name[ 1 ] == ' ') );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    pipe_name ...                                                    */
/*    -------------------------------------------------------------    */
/*    Pipe name to name translation.                                   */
/*---------------------------------------------------------------------*/
static char *
pipe_name( char *pipe_name ) {
   return (pipe_name + 1);
}

/*---------------------------------------------------------------------*/
/*    bool_t                                                           */
/*    directoryp ...                                                   */
/*    -------------------------------------------------------------    */
/*    Is a file a directory?                                           */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF bool_t
directoryp( char *name ) { 
   struct stat buf;

   if( stat( name, &buf ) == -1 )
      return 0;

   return S_ISDIR( buf.st_mode & S_IFMT );
}

/*---------------------------------------------------------------------*/
/*    REGULAR_FILE_NAMEP ...                                           */
/*---------------------------------------------------------------------*/
#define REGULAR_FILE_NAMEP( f ) \
   ((f[ 0 ] != '.') || ((f[ 1 ] != 0) && ((f[ 1 ] != '.') || f[ 2 ] != 0)))

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    directory_to_list ...                                            */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
directory_to_list( char *name ) {
   obj_t res = BNIL;
#ifndef _MSC_VER
   DIR *dir;
   struct dirent *dirent;

   if( (dir = opendir( name )) ) {
      while( (dirent = readdir( dir )) ) {
	 char *fname = dirent->d_name;

	 if( REGULAR_FILE_NAMEP( fname ) ) 
	    res = MAKE_PAIR( string_to_bstring( fname ), res );
      }
      closedir( dir );
   }
#else
   char *const path = (char *)malloc( strlen( name ) + 2 + 1 );

   strcpy( path, name );
   strcat( path, "\\*" );

   {
      WIN32_FIND_DATA find_data;
      HANDLE hSearch = FindFirstFile( path, &find_data );

      if( hSearch != INVALID_HANDLE_VALUE ) {
         BOOL keep_going;

         do {
	    if( REGULAR_FILE_NAMEP( find_data.cFileName ) ) 
               res = MAKE_PAIR( string_to_bstring( find_data.cFileName ), res );
            keep_going= FindNextFile( hSearch, &find_data );
         } while( keep_going );

         FindClose( hSearch );
      }
   }
#endif
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bgl_directory_to_path_list ...                                   */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_directory_to_path_list( char *name, int len, char sep ) {
   obj_t res = BNIL;
#ifndef _MSC_VER
   DIR *dir;
   struct dirent *dirent;

   if( (dir = opendir( name )) ) {
      while( (dirent = readdir( dir )) ) {
	 char *fname = dirent->d_name;

	 if( REGULAR_FILE_NAMEP( fname ) ) { 
	    obj_t bs = make_string_sans_fill( strlen( fname ) + len + 1 );
	    char *s = BSTRING_TO_STRING( bs );

	    strcpy( s, name );
	    s[ len ] = sep;
	    strcpy( s + len + 1, fname );

	    res = MAKE_PAIR( bs, res );
	 }
      }
      closedir( dir );
   }
#else
   char *const path = (char *)malloc( len + 2 + 1 );

   strcpy( path, name );
   strcat( path, "\\*" );

   {
      WIN32_FIND_DATA find_data;
      HANDLE hSearch = FindFirstFile( path, &find_data );

      if( hSearch != INVALID_HANDLE_VALUE ) {
         BOOL keep_going;

         do {
	    char *fname = find_data.cFileName; 
	    if( REGULAR_FILE_NAMEP( fname ) ) { 
	       obj_t bs = make_string_sans_fill( strlen( fname ) + len + 1 );
	       char *s = BSTRING_TO_STRING( bs );
	       strcpy( s, name );
	       s[ len ] = sep;
	       strcpy( s + len + 1, fname );
	       
	       res = MAKE_PAIR( bs, res );
	    }
            keep_going= FindNextFile( hSearch, &find_data );
         } while( keep_going );

         FindClose( hSearch );
      }
   }
#endif
   return res;
}

/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    copyfile ...                                                     */
/*    -------------------------------------------------------------    */
/*    This function is a replacement for sendfile. In particular       */
/*    it is used when the INPUT-PORT is a socket descriptor or when    */
/*    the host system does not support the sendfile system call.       */
/*---------------------------------------------------------------------*/
static long
copyfile( long on, long in, long sz ) {
   long rsz = 0, o;
   if( sz < 0 ) {
#ifdef __GNUC__
      char buf[ default_io_bufsiz + 1 ];
#else
      char *buf = alloca( default_io_bufsiz + 1 );
#endif
      long n, m;

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile on=%d in=%d sz=%d\n", on, in, sz );
#endif
      
loopr:
      while( (n = read( in, buf, default_io_bufsiz )) > 0 ) {
	 o = 0;
loopw:
	 if( (m = write( on, &buf[ o ], n )) < 0 ) {
	    if( errno == EINTR ) {
	       goto loopw;
	    } else {
	       return -1;
	    }
	 }

	 rsz +=m;
	 
	 if( m < n ) {
	    n -= m;
	    o += m;
	    
	    goto loopw;
	 }
      }

      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr;
      }

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile rsz=%d\n", rsz );
#endif
      return rsz;
   } else {
      long n = 0, m;
      long s = (default_io_bufsiz > sz) ? sz : default_io_bufsiz;
#ifdef __GNUC__
      char buf[ s + 1 ];
#else
      char *buf = alloca( s + 1 );
#endif

loopr2:      
      while( (sz > 0) && (n = read( in, buf, s )) > 0 ) {
	 o = 0;
loopw2:
	 if( (m = write( on, &buf[ o ], n )) < 0 ) {
	    if( errno == EINTR )
	       goto loopw2;
	    else 
	       return -1;
	 }

	 rsz += m;
	 sz -= m;
	 
	 if( m < n ) {
	    n -= m;
	    o += m;

	    goto loopw2;
	 }

	 if( sz < s ) s = sz;
      }
      
      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr2;
      } 

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile rsz=%d\n", rsz );
#endif
      return rsz;
   }
}   
   
/*---------------------------------------------------------------------*/
/*    static long                                                      */
/*    copyfile_with_timeout ...                                        */
/*    -------------------------------------------------------------    */
/*    This function is equivalent to COPYFILE. The only difference     */
/*    comes from that is does not inline the call to READ and WRITE.   */
/*---------------------------------------------------------------------*/
static long
copyfile_with_timeout( obj_t op, obj_t ip, long sz ) {
   size_t (*syswrite)() = OUTPUT_PORT( op ).syswrite;
   long (*sysread)() = INPUT_PORT( ip ).sysread;
   long rsz = 0, o;
   if( sz < 0 ) {
#ifdef __GNUC__
      char _buf[ default_io_bufsiz + 1 ];
      char *buf = _buf;
#else
      char *buf = alloca( default_io_bufsiz + 1 );
#endif
      long n;
      size_t m;

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout op=%p ip=%p sz=%d\n", op, ip, sz );
#endif
      
loopr:
      while( (n = sysread( buf, 1, default_io_bufsiz, ip )) > 0 ) {
	 o = 0;
loopw:
	 if( (m = syswrite( &buf[ o ], 1, n, op )) < 0 ) {
	    if( errno == EINTR ) {
	       goto loopw;
	    } else {
	       return -1;
	    }
	 }
	 
	 rsz += m;
	 
	 if( m < n ) {
	    n -= m;
	    o += m;
	    goto loopw;
	 }
      }

      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr;
      }

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout rsz=%d\n", rsz );
#endif
      return rsz;
   } else {
      long n = 0, m;
      long s = (default_io_bufsiz > sz) ? sz : default_io_bufsiz;
#ifdef __GNUC__
      char buf[ s + 1 ];
#else
      char *buf = alloca( s + 1 );
#endif

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout op=%p ip=%p sz=%d\n", op, ip, sz );
#endif
loopr2:      
      while( (sz > 0) && (n = sysread( buf, 1, s, ip )) > 0 ) {
	 o = 0;
loopw2:
	 if( (m = syswrite( &buf[ o ], 1, n, op )) < 0 ) {
	    if( errno == EINTR )
	       goto loopw2;
	    else 
	       return -1;
	 }

	 rsz += m;
	 sz -= m;
	 
	 if( m < n ) {
	    n -= m;
	    o += m;
	    goto loopw2;
	 }

	 if( sz < s ) s = sz;
      }

      if( (n < 0) && (errno == EINTR) ) {
	 goto loopr2;
      } 

#ifdef DEBUG_SENDCHARS
      fprintf( stderr, "copyfile_with_timeout rsz=%d\n", rsz );
#endif
      return rsz;
   }
}   

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    gc_sendfile ...                                                  */
/*---------------------------------------------------------------------*/
#if( BGL_GC_HAVE_DO_BLOCKING )
static void
gc_sendfile( struct sendfile_info_t *si ) {
#if DEBUG_SENDCHARS
   fprintf( stderr, "gc_sendfile out=%p in=%p offset%d sz=%d\n",
	    si->out, si->in, si->offset, si->sz );
#endif   
   si->res = sendfile( si->out, si->in, si->offset, si->sz );
}
#endif

/*---------------------------------------------------------------------*/
/*    bgl_sendchars ...                                                */
/*    -------------------------------------------------------------    */
/*    uses sendfile to "copy" the input-port to the output-port        */
/*    flushes output-port!                                             */
/*---------------------------------------------------------------------*/
obj_t
bgl_sendchars( obj_t ip, obj_t op, long sz, long offset ) {
#define inp INPUT_PORT( ip )
#define outp OUTPUT_PORT( op )
   long dsz;
   long ws = 0;
   struct stat in;
   struct stat out;
   long n;

   switch( (long)(inp.port.kindof) ) {
      case (long)KINDOF_STRING:
      case (long)KINDOF_CLOSED:
      case (long)KINDOF_PROCEDURE:
      case (long)KINDOF_GZIP:
	 return BFALSE;
   }

   if( OUTPUT_STRING_PORTP( op ) )
       return BFALSE;
       
   switch( (long)(outp.port.kindof) ) {
      case (long)KINDOF_CLOSED:
      case (long)KINDOF_PROCEDURE:
	 return BFALSE;
   }

#ifdef DEBUG_SENDCHARS   
   fprintf( stderr, "bgl_sendchars.1: sz=%d offset=%d filepos=%d\n", sz, offset,
	    inp.filepos );
#endif   
   
   if( offset >= 0 ) {
      bgl_input_port_seek( ip, offset );
   } else {
      dsz = inp.bufpos - inp.matchstop - 1;
   
#ifdef DEBUG_SENDCHARS   
      fprintf( stderr, "bgl_sendchars.2: sz=%d offset=%d dsz=%d\n", sz, offset, dsz );
#endif      
   
      outp.sysflush( op );

      if( dsz > 0 ) {
	 /* flush the buffer when it contains characters */
	 size_t w;

	 ws = ((sz > 0) && (dsz > sz)) ? sz : dsz;
	 w = outp.syswrite( &(inp.buffer[ inp.matchstop ]), 1, ws, op );

#ifdef DEBUG_SENDCHARS   
	 fprintf( stderr, "bgl_sendchars.3: sz=%d offset=%d w=%d ws=%d\n", sz, offset, w, ws );
#endif	 
	 inp.matchstop += (long)w;
	 inp.forward = inp.matchstop;

	 if( w < ws ) {
	    if( ferror( (FILE *)outp.port.stream ) || outp.sysflush( op ) ) {
	       C_SYSTEM_FAILURE( bglerror( errno ),
				 "send-chars",
				 strerror( errno ),
				 MAKE_PAIR( ip, op ) );
	       return BINT( w );
	    }
	 }

	 outp.sysflush( op );

	 if( sz > 0 ) {
	    if( dsz >= sz ) {
#ifdef DEBUG_SENDCHARS   
	       fprintf( stderr, "bgl_sendchars.4: RETURN ws=%d\n", ws );
#endif	       
	       return BINT( ws );
	    }

	    sz -= ws;
	 }
      }
   }
     
#if( BGL_HAVE_SENDFILE )
   if( 
      /* Some operating systems (such as Linux 2.6.10) are demanding */
      /* on the input and output ports. These requirements are set   */
      /* in the configuration files and used to determine what has   */
      /* to be checked before invoking the actual sendfile sys call  */
#  if( BGL_SENDFILE_REQUIRE_INPUT_FILE )      
       (inp.port.kindof != KINDOF_FILE)
       || fstat( fileno( (FILE *)inp.port.stream ), &in )
       || !S_ISREG( in.st_mode )
#  else
       0
#  endif
#  if( BGL_SENDFILE_REQUIRE_OUTPUT_SOCKET )
       || (outp.port.kindof != KINDOF_FILE)
       || fstat( fileno( (FILE *)outp.port.stream ), &out )
       || ((out.st_mode & S_IFSOCK) != S_IFSOCK)
#  endif       
       || 0 ) {
#endif /* BGL_HAVE_SENDFILE */
      if( sz != 0 ) {
	 if( PORT( ip ).timeout || PORT( op ).timeout ) {
	    n = copyfile_with_timeout( op, ip, sz );
	 } else {
	    n = copyfile( fileno( (FILE *)outp.port.stream ),
			  fileno( (FILE *)inp.port.stream ),
			  sz );
	 }
      }
      if( n < 0 ) {
	 C_SYSTEM_FAILURE( bglerror( errno ), 
			   "send-chars",
			   strerror( errno ),
			   MAKE_PAIR( ip, op ) );
	 return BINT( 0 );
      }
#if( BGL_HAVE_SENDFILE )
   } else {
      if( sz == -1 ) sz = in.st_size;

      if( sz != 0 ) {
#if( BGL_GC_HAVE_BLOCKING )
	 bgl_gc_start_blocking();

	 n = sendfile( fileno( (FILE *)outp.port.stream ),
		       fileno( (FILE*)inp.port.stream ),
		       (offset > 0 ? (off_t *)(&offset) : 0),
		       sz );
	 
	 bgl_gc_stop_blocking();
#else
#  if( BGL_GC_HAVE_DO_BLOCKING )
	 {
	    struct sendfile_info_t si;
	    si.out = fileno( (FILE *)outp.port.stream );
	    si.in = fileno( (FILE *)inp.port.stream );
	    si.offset = (offset > 0 ? (off_t *)(&offset) : 0);
	    si.sz = sz;
	 
#ifdef DEBUG_SENDCHARS   
	    fprintf( stderr, "bgl_sendchars.5: sz=%d offset=%d\n", sz, offset );
#endif	    
	    bgl_gc_do_blocking( &gc_sendfile, &si );

	    n = si.res;
#ifdef DEBUG_SENDCHARS   
	    fprintf( stderr, "bgl_sendchars.6: n=%d\n", n );
#endif	    
	 }
      }
#  else
      -> error BGL_GC_HAVE_BLOCKING or BGL_GC_HAVE_DO_BLOCKING required
#  endif       
#endif /* BGL_HAVE_BLOCKING */
      if( n < 0 ) {
	 C_SYSTEM_FAILURE( bglerror( errno ),
			   "send-chars",
			   strerror( errno ),
			   MAKE_PAIR( ip, op ) );
      }
   }
   if( offset > 0 ) {
      if( INPUT_PORT_ON_FILEP( ip ) )
	 fseek( (FILE *)inp.port.stream, offset + n + ws, SEEK_SET );
   }
#endif /* BGL_HAVE_SENDFILE */
   
   inp.filepos += n + ws;

   return BINT( n + ws );
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    strwrite ...                                                     */
/*---------------------------------------------------------------------*/
static size_t
strwrite( void *ptr, size_t size, size_t nmemb, obj_t p ) {
   long len = (long)(size == 1 ? nmemb : size * nmemb);
   long offset = OUTPUT_STRING_PORT( p ).offset;

   while( (OUTPUT_STRING_PORT( p ).offset+len) > OUTPUT_STRING_PORT( p ).size )
      bgl_strport_grow( p );

   memcpy( &(OUTPUT_STRING_PORT(p).buffer[ offset ] ), ptr, len );

   OUTPUT_STRING_PORT( p ).offset = offset + len;

   return len;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    strputc ...                                                      */
/*---------------------------------------------------------------------*/
static int
strputc( int c, obj_t p ) {
   long offset = OUTPUT_STRING_PORT( p ).offset;
   
   if( END_OF_STRING_PORTP( p ) )
      bgl_strport_grow( p );

   OUTPUT_STRING_PORT( p ).buffer[ offset ] = c;
   OUTPUT_STRING_PORT( p ).offset = offset + 1;

   return c;
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    procwrite ...                                                    */
/*---------------------------------------------------------------------*/
static size_t
procwrite( void *ptr, size_t size, size_t nmemb, obj_t p ) {
   obj_t proc = OUTPUT_PROCEDURE_PORT( p ).proc;
   obj_t buf = OUTPUT_PROCEDURE_PORT( p ).buffer;
   long sz = size * nmemb;

   if( sz > STRING_LENGTH( buf ) ) {
      OUTPUT_PROCEDURE_PORT( p ).buflen = sz + 1;
      buf = OUTPUT_PROCEDURE_PORT( p ).buffer = make_string( sz + 1 );
   }
      
   memcpy( &STRING_REF( buf, 0 ), ptr, sz );
   STRING_SET( buf, sz, '\0' );
   STRING_LENGTH( buf ) = sz;
   
   PROCEDURE_ENTRY( proc )( proc, buf, BEOA );
   return sz;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    procputc ...                                                     */
/*---------------------------------------------------------------------*/
static int
procputc( int c, obj_t p ) {
   obj_t proc = OUTPUT_PROCEDURE_PORT( p ).proc;
   obj_t buf = OUTPUT_PROCEDURE_PORT( p ).buffer;
   
   STRING_SET( buf, 0, c );
   STRING_SET( buf, 1, '\0' );
   STRING_LENGTH( buf ) = 1;
   
   PROCEDURE_ENTRY( proc )( proc, buf, BEOA );
   return c;
}

/*---------------------------------------------------------------------*/
/*    static size_t                                                    */
/*    closedwrite ...                                                  */
/*---------------------------------------------------------------------*/
static size_t
closedwrite( void *ptr, size_t size, size_t nmemb, obj_t port ) {
   C_SYSTEM_FAILURE( BGL_IO_CLOSED_ERROR,
		     "write",
		     "Illegal write to a closed port",
		     port );
   return -1;
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    closedputc ...                                                   */
/*---------------------------------------------------------------------*/
static int
closedputc( int c, obj_t port ) {
   C_SYSTEM_FAILURE( BGL_IO_CLOSED_ERROR,
		     "write",
		     "Illegal write to a closed port",
		     port );
   return -1;
}

/*---------------------------------------------------------------------*/
/*    BGL_RUNTIME_DEF obj_t                                            */
/*    bgl_password ...                                                 */
/*---------------------------------------------------------------------*/
BGL_RUNTIME_DEF obj_t
bgl_password( char *prompt ) {
   int max_len = 80;
   char *s = alloca( max_len + 1 );
   int i = 0;
   int c;
   FILE *tty = fopen( "/dev/tty", "w" );
   FILE *out = tty ? tty : stderr;
   
#if HAVE_TERMIO || HAVE_TERMIOS
   struct termios t;
   tcflag_t lflag;

   fputs( prompt, out );
   fflush( out );

   /* set the terminal in no echo mode */
   tcgetattr( 0, &t );
   lflag = t.c_lflag;
   t.c_lflag &= ~(ICANON|ECHO);
   t.c_cc[ VMIN ] = 1;
   t.c_cc[ VTIME ] = 0;
   tcsetattr( 0, TCSANOW, &t );
#endif

   while( (c = getchar()) != '\n' ) {
      if( i == max_len ) {
	 char *ns = alloca( max_len * 2 );
	 memcpy( ns, s, max_len );
	 max_len *= 2;
	 s = ns;
      }
      
      s[ i++ ] = c;
      putc( '*', out );
      fflush( out );
   }

   s[ i ] = 0;
   
#if HAVE_TERMIO || HAVE_TERMIOS
   /* restore the terminal */
   t.c_lflag = lflag;
   tcsetattr( 0, TCSANOW, &t );
#endif

   putc( '\n', out );
   fflush( out );
   
   if( tty ) fclose( tty );
   
   return string_to_bstring_len( s, i );
}
