/*=====================================================================*/
/*    /opt/tmp-23apr07/bigloo/runtime/Jlib/server_socket.java          */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Tue Dec  5 10:53:03 2000                          */
/*    Last change :  Mon Apr 23 07:38:37 2007 (serrano)                */
/*    Copyright   :  2000-07 Manuel Serrano                            */
/*    -------------------------------------------------------------    */
/*    The Server Socket implementation for the JVM back-end.           */
/*=====================================================================*/
package bigloo;

import java.io.*;
import java.net.*;

/*---------------------------------------------------------------------*/
/*    SOCKET ...                                                       */
/*---------------------------------------------------------------------*/
public class server_socket extends socket {
   /*--- private fields ---------------------------------------------------*/
   protected ServerSocket server_socket;
   protected Socket client_socket;
   protected boolean blocking;

   /*--- constructors ----------------------------------------------------*/
   public server_socket( Object name, final int port ) {
      super();

      try {
         String n = (name == bbool.faux) ? null : new String((byte[])name);
         server_socket = JDK.makeServerSocket(n, port);
      } catch (final IOException _) {
	 socket_error( "make-server-socket",
		       "cannot create socket",
		       new bint( port ) );
      }
   }

   protected server_socket() {
   }

   private server_socket( final server_socket templ ) {
      super();
      server_socket= templ.server_socket;
      client_socket= templ.client_socket;
      blocking= templ.blocking;
   }

   /*--- public methods --------------------------------------------------*/
   public Object HOSTNAME() {
      if (client_socket != null)
	 return client_socket.getInetAddress().getHostName().getBytes();
      else
	 return bigloo.foreign.BFALSE;
   }

   public Object HOSTIP() {
      if (client_socket != null)
	 return client_socket.getInetAddress().getHostAddress().getBytes();
      else
	 return bigloo.foreign.BFALSE;
   }

   public byte[] local_addr() {
      return server_socket.getInetAddress().getHostAddress().getBytes();
   }

   public client_socket accept( boolean blocking, boolean errp )
      throws IOException, SecurityException {
      // WARNING this blocking parameter is not called blocking in C
      // it is called buffered, and C does not do non-blocking [yet], so
      // we don't support it [yet] (but we can)
      final Socket accepted_socket = JDK.accept(server_socket, true);
      if( accepted_socket == null )
	 throw new IOException("Nothing to accept");

      return new client_socket( accepted_socket, blocking );
   }

   public Object shutdown( final boolean  close_socket ) throws IOException {
      if (client_socket != null)
	 try {
	    client_socket.shutdownOutput();
	    client_socket.shutdownInput();
	 } catch (Exception _) {
	 }

      if (close_socket)
	 close();

      return bigloo.foreign.BUNSPEC;
   }

   public Object close() throws IOException {
      server_socket.close();
      down= true;
      return bigloo.foreign.BUNSPEC;
   }

   public int PORT() {
      return server_socket.getLocalPort();
   }

   public void write( final output_port  p ) {
      p.write( "#<socket:" + server_socket.getInetAddress() 
	       + ":" + PORT() + ">" );
   }
  
   public ServerSocket getSocket() {
      return server_socket;
   }
}
